import os
import unittest
import requests_mock
from unittest import mock
from test.utils import assert_matching_tables
from parsons.bloomerang.bloomerang import Bloomerang
from parsons.etl import Table

from test.test_bloomerang.test_data import ENV_PARAMETERS, ID, TEST_DELETE, \
    TEST_CREATE_CONSTITUENT, TEST_GET_CONSTITUENT, TEST_GET_CONSTITUENTS, \
    TEST_CREATE_TRANSACTION, TEST_GET_TRANSACTION, TEST_GET_TRANSACTIONS, \
    TEST_CREATE_INTERACTION, TEST_GET_INTERACTION, TEST_GET_INTERACTIONS


class TestBloomerang(unittest.TestCase):

    def setUp(self):
        self.bloomerang = Bloomerang(api_key='test_key')

    @mock.patch.dict(os.environ, ENV_PARAMETERS)
    def test_init_env(self):
        bloomerang = Bloomerang()
        self.assertEqual(bloomerang.api_key, 'env_api_key')
        self.assertEqual(bloomerang.client_id, 'env_client_id')
        self.assertEqual(bloomerang.client_secret, 'env_client_secret')

    @requests_mock.Mocker()
    def test_authentication(self, m):
        # API key
        bloomerang = Bloomerang(api_key='my_key')
        self.assertEqual(bloomerang.conn.headers['X-API-KEY'], 'my_key')

        # OAuth2
        m.post(url=bloomerang.uri_auth, json={'code': 'my_auth_code'})
        m.post(url=bloomerang.uri + 'oauth/token', json={'access_token': 'my_access_token'})
        bloomerang = Bloomerang(client_id='my_id', client_secret='my_secret')
        self.assertEqual(bloomerang.authorization_code, 'my_auth_code')
        self.assertEqual(bloomerang.access_token, 'my_access_token')
        self.assertEqual(bloomerang.conn.headers['Authorization'], 'Bearer my_access_token')

    def test_base_endpoint(self):
        url = self.bloomerang._base_endpoint('constituent')
        self.assertEqual(url, 'https://api.bloomerang.co/v2/constituent/')

        url = self.bloomerang._base_endpoint('constituent', 1234)
        self.assertEqual(url, 'https://api.bloomerang.co/v2/constituent/1234/')

        url = self.bloomerang._base_endpoint('constituent', '1234')
        self.assertEqual(url, 'https://api.bloomerang.co/v2/constituent/1234/')

    @requests_mock.Mocker()
    def test_create_constituent(self, m):
        m.post(f'{self.bloomerang.uri}constituent/', json=TEST_CREATE_CONSTITUENT)
        self.assertEqual(self.bloomerang.create_constituent(), TEST_CREATE_CONSTITUENT)

    @requests_mock.Mocker()
    def test_update_constituent(self, m):
        m.put(f'{self.bloomerang.uri}constituent/{ID}/', json=TEST_CREATE_CONSTITUENT)
        self.assertEqual(self.bloomerang.update_constituent(ID), TEST_CREATE_CONSTITUENT)

    @requests_mock.Mocker()
    def test_get_constituent(self, m):
        m.get(f'{self.bloomerang.uri}constituent/{ID}/', json=TEST_GET_CONSTITUENT)
        self.assertEqual(self.bloomerang.get_constituent(ID), TEST_GET_CONSTITUENT)

    @requests_mock.Mocker()
    def test_delete_constituent(self, m):
        m.delete(f'{self.bloomerang.uri}constituent/{ID}/', json=TEST_DELETE)
        self.assertEqual(self.bloomerang.delete_constituent(ID), TEST_DELETE)

    @requests_mock.Mocker()
    def test_get_constituents(self, m):
        m.get(f'{self.bloomerang.uri}constituents/?skip=0&take=50', json=TEST_GET_CONSTITUENTS)
        assert_matching_tables(self.bloomerang.get_constituents(),
                               Table(TEST_GET_CONSTITUENTS['Results']))

    @requests_mock.Mocker()
    def test_create_transaction(self, m):
        m.post(f'{self.bloomerang.uri}transaction/', json=TEST_CREATE_TRANSACTION)
        self.assertEqual(self.bloomerang.create_transaction(), TEST_CREATE_TRANSACTION)

    @requests_mock.Mocker()
    def test_update_transaction(self, m):
        m.put(f'{self.bloomerang.uri}transaction/{ID}/', json=TEST_CREATE_TRANSACTION)
        self.assertEqual(self.bloomerang.update_transaction(ID), TEST_CREATE_TRANSACTION)

    @requests_mock.Mocker()
    def test_get_transaction(self, m):
        m.get(f'{self.bloomerang.uri}transaction/{ID}/', json=TEST_GET_TRANSACTION)
        self.assertEqual(self.bloomerang.get_transaction(ID), TEST_GET_TRANSACTION)

    @requests_mock.Mocker()
    def test_delete_transaction(self, m):
        m.delete(f'{self.bloomerang.uri}transaction/{ID}/', json=TEST_DELETE)
        self.assertEqual(self.bloomerang.delete_transaction(ID), TEST_DELETE)

    @requests_mock.Mocker()
    def test_get_transactions(self, m):
        m.get(f'{self.bloomerang.uri}transactions/?skip=0&take=50', json=TEST_GET_TRANSACTIONS)
        assert_matching_tables(self.bloomerang.get_transactions(),
                               Table(TEST_GET_TRANSACTIONS['Results']))

    @requests_mock.Mocker()
    def test_get_transaction_designation(self, m):
        m.get(f'{self.bloomerang.uri}transaction/designation/{ID}/', json=TEST_GET_TRANSACTION)
        self.assertEqual(self.bloomerang.get_transaction_designation(ID), TEST_GET_TRANSACTION)

    @requests_mock.Mocker()
    def test_get_transaction_designations(self, m):
        m.get(f'{self.bloomerang.uri}transactions/designations/?skip=0&take=50',
              json=TEST_GET_TRANSACTIONS)
        assert_matching_tables(self.bloomerang.get_transaction_designations(),
                               Table(TEST_GET_TRANSACTIONS['Results']))

    @requests_mock.Mocker()
    def test_create_interaction(self, m):
        m.post(f'{self.bloomerang.uri}interaction/', json=TEST_CREATE_INTERACTION)
        self.assertEqual(self.bloomerang.create_interaction(), TEST_CREATE_INTERACTION)

    @requests_mock.Mocker()
    def test_update_interaction(self, m):
        m.put(f'{self.bloomerang.uri}interaction/{ID}/', json=TEST_CREATE_INTERACTION)
        self.assertEqual(self.bloomerang.update_interaction(ID), TEST_CREATE_INTERACTION)

    @requests_mock.Mocker()
    def test_get_interaction(self, m):
        m.get(f'{self.bloomerang.uri}interaction/{ID}/', json=TEST_GET_INTERACTION)
        self.assertEqual(self.bloomerang.get_interaction(ID), TEST_GET_INTERACTION)

    @requests_mock.Mocker()
    def test_delete_interaction(self, m):
        m.delete(f'{self.bloomerang.uri}interaction/{ID}/', json=TEST_DELETE)
        self.assertEqual(self.bloomerang.delete_interaction(ID), TEST_DELETE)

    @requests_mock.Mocker()
    def test_get_interactions(self, m):
        m.get(f'{self.bloomerang.uri}interactions/?skip=0&take=50', json=TEST_GET_INTERACTIONS)
        assert_matching_tables(self.bloomerang.get_interactions(),
                               Table(TEST_GET_INTERACTIONS['Results']))



from parsons.pdi.flag_ids import FlagIDs
from parsons.pdi.flags import Flags
from parsons.pdi.universes import Universes
from parsons.pdi.questions import Questions
from parsons.pdi.acquisition_types import AcquisitionTypes

from parsons import Table
from parsons.utilities import check_env

from datetime import datetime, timezone
from dateutil.parser import parse
from json.decoder import JSONDecodeError
import logging
import requests


logger = logging.getLogger(__name__)


class PDI(FlagIDs, Universes, Questions, AcquisitionTypes, Flags):

    def __init__(self, username=None, password=None, api_token=None,
                 qa_url=False):
        """
        Instantiate the PDI class

        `Args:`
            username: str
                The username for a PDI account. Can be passed as arguement or
                can be set as `PDI_USERNAME` environment variable.
            password: str
                The password for a PDI account. Can be passed as arguement or
                can be set as `PDI_PASSWORD` environment variable.
            api_token: str
                The api_token for a PDI account. Can be passed as arguement or
                can be set as `PDI_API_TOKEN` environment variable.
            qa_url: bool
                Defaults to False. If True, requests will be made to a sandbox
                account. NOTE: This requires separate qa credentials and api
                token.
        """
        if qa_url:
            self.base_url = "https://apiqa.bluevote.com"
        else:
            self.base_url = "https://api.bluevote.com"

        self.username = check_env.check('PDI_USERNAME', username)
        self.password = check_env.check('PDI_PASSWORD', password)
        self.api_token = check_env.check('PDI_API_TOKEN', api_token)

        super().__init__()

        self._get_session_token()

    def _get_session_token(self):
        headers = {
            "Content-Type": "application/json",
        }
        login = {
            "Username": self.username,
            "Password": self.password,
            "ApiToken": self.api_token,
        }
        res = requests.post(
            f"{self.base_url}/sessions",
            json=login,
            headers=headers)

        logger.debug(f"{res.status_code} - {res.url}")
        res.raise_for_status()

        # status_code == 200
        data = res.json()
        self.session_token = data["AccessToken"]
        self.session_exp = parse(data["ExpirationDate"])

    def _clean_dict(self, dct):
        if isinstance(dct, list):
            return [self._clean_dict(obj) for obj in dct]

        if isinstance(dct, dict):
            return {k: v for k, v in dct.items() if v is not None}

        return dct

    def _request(self, url, req_type='GET', post_data=None, args=None,
                 limit=None):
        # Make sure to have a current token before we make another request
        now = datetime.now(timezone.utc)
        if now > self.session_exp:
            self._get_session_token()

        # Based on PDI docs
        # https://api.bluevote.com/docs/index
        LIMIT_MAX = 2000

        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.session_token}",
        }

        request_fn = {
            "GET": requests.get,
            "POST": requests.post,
            "PUT": requests.put,
            "DELETE": requests.delete,
        }

        if limit and limit <= LIMIT_MAX:
            args = args or {}
            args["limit"] = limit

        args = self._clean_dict(args) if args else args
        post_data = self._clean_dict(post_data) if post_data else post_data

        res = request_fn[req_type](
            url, headers=headers, json=post_data, params=args)

        logger.debug(f"{res.url} - {res.status_code}")
        logger.debug(res.request.body)

        res.raise_for_status()

        if not res.text:
            return None

        logger.debug(res.text)

        try:
            res_json = res.json()
        except JSONDecodeError:
            res_json = None

        if "data" not in res_json:
            return res_json

        total_count = (0 if "totalCount" not in res_json
                       else res_json["totalCount"])
        data = res_json["data"]

        if not limit:
            # We don't have a limit, so let's get everything
            # Start a page 2 since we already go page 1
            cursor = 2
            while len(data) < total_count:
                args = args or {}
                args["cursor"] = cursor
                args["limit"] = LIMIT_MAX
                res = request_fn[req_type](
                    url, headers=headers, json=post_data, params=args)

                data.extend(res.json()["data"])

                cursor += 1

            return Table(data)

        else:
            total_need = min(limit, total_count)

            cursor = 2
            while len(data) < total_need:
                args = args or {}
                args["cursor"] = cursor
                args["limit"] = min(LIMIT_MAX, total_need - len(data))
                res = request_fn[req_type](
                    url, headers=headers, json=post_data, params=args)

                data.extend(res.json()["data"])

                cursor += 1

            return Table(data)



from parsons.databases.database.database import DatabaseCreateStatement
import parsons.databases.postgres.constants as consts

import petl
import logging

logger = logging.getLogger(__name__)


class PostgresCreateStatement(DatabaseCreateStatement):

    def __init__(self):
        super().__init__()

        self.COL_NAME_MAX_LEN = consts.COL_NAME_MAX_LEN
        self.REPLACE_CHARS = consts.REPLACE_CHARS

        # Postgres doesn't have a medium int
        self.MEDIUMINT = self.INT

        # Currently py floats are coded as Postgres decimals
        self.FLOAT = consts.DECIMAL

        # Max length of a Redshift VARCHAR column
        self.VARCHAR_MAX = consts.VARCHAR_MAX

        # List of varchar lengths to use for columns -- this list needs to be in order from
        # smallest to largest
        self.VARCHAR_STEPS = consts.VARCHAR_STEPS

    # the default behavior is f"{col}_"
    def _rename_reserved_word(self, col, index):
        """Return the renamed column.

        `Args`:
            col: str
                The column to rename.
            index: int
                (Optional) The index of the column.
        `Returns`:
            str
                The rename column.
        """
        return f"col_{index}"

    def create_statement(self, tbl, table_name, padding=None, distkey=None, sortkey=None,
                         varchar_max=None, varchar_truncate=True, columntypes=None,
                         strict_length=True):
        # Generate a table create statement. Distkeys and sortkeys are only used by
        # Redshift and should not be passed when generating a create statement for
        # Postgres.

        if tbl.num_rows == 0:
            raise ValueError('Table is empty. Must have 1 or more rows.')

        # Validate and rename column names if needed
        tbl.table = petl.setheader(tbl.table, self.column_name_validate(tbl.columns))

        mapping = self.generate_data_types(tbl)

        if padding:
            mapping['longest'] = self.vc_padding(mapping, padding)
        elif not strict_length:
            mapping['longest'] = self.vc_step(mapping)

        if varchar_max:
            mapping['longest'] = self.vc_max(mapping, varchar_max)

        if varchar_truncate:
            mapping['longest'] = self.vc_trunc(mapping)

        mapping['longest'] = self.vc_validate(mapping)

        # Add any provided column type overrides
        if columntypes:
            for i in range(len(mapping['headers'])):
                col = mapping['headers'][i]
                if columntypes.get(col):
                    mapping['type_list'][i] = columntypes[col]

        # Enclose in quotes
        mapping['headers'] = [f'"{h}"'for h in mapping['headers']]

        return self.create_sql(table_name, mapping, distkey=distkey, sortkey=sortkey)

    # This is for backwards compatability
    def data_type(self, val, current_type):
        return self.detect_data_type(val, current_type)

    # This is for backwards compatability
    def is_valid_integer(self, val):
        return self.is_valid_sql_num(val)

    def generate_data_types(self, table):
        # Generate column data types

        longest, type_list = [], []

        cont = petl.records(table.table)

        # Populate empty values for the columns
        for col in table.columns:
            longest.append(0)
            type_list.append('')

        for row in cont:
            for i in range(len(row)):
                # NA is the csv null value
                if type_list[i] == 'varchar' or row[i] in ['NA', '']:
                    pass
                else:
                    var_type = self.data_type(row[i], type_list[i])
                    type_list[i] = var_type

                # Calculate width
                width = len(str(row[i]).encode('utf-8'))
                if width > longest[i]:
                    longest[i] = width

        # In L140 'NA' and '' will be skipped
        # If the entire column is either one of those (or a mix of the two)
        # the type will be empty.
        # Fill with a default varchar
        type_list = [typ or 'varchar' for typ in type_list]

        return {'longest': longest,
                'headers': table.columns,
                'type_list': type_list}

    def vc_padding(self, mapping, padding):
        # Pad the width of a varchar column

        return [int(c + (c * padding)) for c in mapping['longest']]

    def vc_step(self, mapping):
        return [self.round_longest(c) for c in mapping['longest']]

    def vc_max(self, mapping, columns):
        # Set the varchar width of a column to the maximum

        for c in columns:

            try:
                idx = mapping['headers'].index(c)
                mapping['longest'][idx] = self.VARCHAR_MAX

            except KeyError as error:
                logger.error('Could not find column name provided.')
                raise error

        return mapping['longest']

    def vc_trunc(self, mapping):

        return [self.VARCHAR_MAX if c > self.VARCHAR_MAX else c for c in mapping['longest']]

    def vc_validate(self, mapping):

        return [1 if c == 0 else c for c in mapping['longest']]

    def create_sql(self, table_name, mapping, distkey=None, sortkey=None):
        # Generate the sql to create the table

        statement = f'create table {table_name} ('

        for i in range(len(mapping['headers'])):
            if mapping['type_list'][i] == 'varchar':
                statement = (statement + '\n  {} varchar({}),').format(str(mapping['headers'][i])
                                                                       .lower(),
                                                                       str(mapping['longest'][i]))
            else:
                statement = (statement + '\n  ' + '{} {}' + ',').format(str(mapping['headers'][i])
                                                                        .lower(),
                                                                        mapping['type_list'][i])

        statement = statement[:-1] + ') '

        if distkey:
            statement += f'\ndistkey({distkey}) '

        if sortkey:
            statement += f'\nsortkey({sortkey})'

        statement += ';'

        return statement

    # This is for backwards compatability
    def column_name_validate(self, columns):
        return self.format_columns(columns, col_prefix="col_")

    @staticmethod
    def round_longest(longest):
        # Find the value that will work best to fit our longest column value
        for step in consts.VARCHAR_STEPS:
            # Make sure we have padding
            if longest < step / 2:
                return step

        return consts.VARCHAR_MAX



test_campaigns = {
    "campaigns": [
        {
            "id": "abc",
            "web_id": 123,
            "type": "regular",
            "create_time": "2019-10-18T02:35:05+00:00",
            "archive_url": "http://example.com/sample-campaign-1",
            "long_archive_url": "https://mailchi.mp/abc/sample-campaign-1",
            "status": "sent",
            "emails_sent": 145,
            "send_time": "2019-10-18T14:15:00+00:00",
            "content_type": "template",
            "needs_block_refresh": False,
            "resendable": True,
            "recipients": {
                "list_id": "zyx",
                "list_is_active": True,
                "list_name": "Support Our Candidate List 1",
                "segment_text": "",
                "recipient_count": 145
            },
            "settings": {
                "subject_line": "Sample Campaign 1",
                "preview_text": "This is a sample campaign.",
                "title": "Sample Campaign Donation Ask",
                "from_name": "Our Candidate",
                "reply_to": "our_candidate@example.com",
                "use_conversation": False,
                "to_name": "*|FNAME|* *|LNAME|*",
                "folder_id": "",
                "authenticate": True,
                "auto_footer": False,
                "inline_css": False,
                "auto_tweet": False,
                "fb_comments": True,
                "timewarp": False,
                "template_id": 12345,
                "drag_and_drop": True
            },
            "tracking": {
                "opens": True,
                "html_clicks": True,
                "text_clicks": False,
                "goal_tracking": False,
                "ecomm360": False,
                "google_analytics": "",
                "clicktale": ""
            },
            "report_summary": {
                "opens": 48,
                "unique_opens": 34,
                "open_rate": 0.23776223776223776,
                "clicks": 1,
                "subscriber_clicks": 1,
                "click_rate": 0.006993006993006993,
                "ecommerce": {
                    "total_orders": 0,
                    "total_spent": 0,
                    "total_revenue": 0
                }
            },
            "delivery_status": {
                "enabled": False
            },
            "_links": []
        },
        {
            "id": "def",
            "web_id": 456,
            "type": "regular",
            "create_time": "2019-05-29T11:46:41+00:00",
            "archive_url": "http://example.com/sample-campaign-2",
            "long_archive_url": "https://mailchi.mp/abc/sample-campaign-2",
            "status": "sent",
            "emails_sent": 87,
            "send_time": "2019-05-29T21:18:15+00:00",
            "content_type": "template",
            "needs_block_refresh": False,
            "resendable": True,
            "recipients": {
                "list_id": "wvu",
                "list_is_active": True,
                "list_name": "Support Our Candidate List 2",
                "segment_text": "",
                "recipient_count": 87
            },
            "settings": {
                "subject_line": "Sample Campaign 2",
                "preview_text": "This is another sample campaign.",
                "title": "Sample Campaign 2 Donation Ask",
                "from_name": "Our Candidate",
                "reply_to": "our_candidate@example.com",
                "use_conversation": False,
                "to_name": "*|FNAME|* *|LNAME|*",
                "folder_id": "",
                "authenticate": True,
                "auto_footer": False,
                "inline_css": False,
                "auto_tweet": False,
                "fb_comments": True,
                "timewarp": False,
                "template_id": 67890,
                "drag_and_drop": True
            },
            "tracking": {
                "opens": True,
                "html_clicks": True,
                "text_clicks": False,
                "goal_tracking": False,
                "ecomm360": False,
                "google_analytics": "",
                "clicktale": ""
            },
            "report_summary": {
                "opens": 108,
                "unique_opens": 48,
                "open_rate": 0.5647058823529412,
                "clicks": 25,
                "subscriber_clicks": 14,
                "click_rate": 0.16470588235294117,
                "ecommerce": {
                    "total_orders": 0,
                    "total_spent": 0,
                    "total_revenue": 0
                }
            },
            "delivery_status": {
                "enabled": False
            },
            "_links": []
        }]}

test_lists = {
    "lists": [
        {
            "id": "zyx",
            "web_id": 98765,
            "name": "Support Our Candidate List 1",
            "contact": {
                "company": "Support Our Candidate",
                "address1": "123 Main Street",
                "address2": "",
                "city": "Townsville",
                "state": "OH",
                "zip": "43358",
                "country": "US",
                "phone": ""
            },
            "permission_reminder": (
                "You are receiving this email because you signed up at an event, while being "
                "canvassed, or on our website."),
            "use_archive_bar": True,
            "campaign_defaults": {
                "from_name": "Our Candidate",
                "from_email": "our_candidate@example.com",
                "subject": "",
                "language": "en"
            },
            "notify_on_subscribe": "",
            "notify_on_unsubscribe": "",
            "date_created": "2019-03-25T22:55:44+00:00",
            "list_rating": 3,
            "email_type_option": False,
            "subscribe_url_short": "http://example.com/sample-subscribe_url_2",
            "subscribe_url_long": "https://mailchi.mp/zyx/sample-subscribe-url-2",
            "beamer_address": "us00-sample-2@inbound.mailchimp.com",
            "visibility": "pub",
            "double_optin": False,
            "has_welcome": True,
            "marketing_permissions": False,
            "modules": [],
            "stats": {
                "member_count": 140,
                "unsubscribe_count": 8,
                "cleaned_count": 16,
                "member_count_since_send": 0,
                "unsubscribe_count_since_send": 1,
                "cleaned_count_since_send": 0,
                "campaign_count": 21,
                "campaign_last_sent": "2020-01-06T01:54:32+00:00",
                "merge_field_count": 5,
                "avg_sub_rate": 0,
                "avg_unsub_rate": 1,
                "target_sub_rate": 3,
                "open_rate": 38.40236686390532,
                "click_rate": 4.016786570743405,
                "last_sub_date": "2019-09-24T01:07:56+00:00",
                "last_unsub_date": "2020-01-06T01:55:02+00:00"
            },
            "_links": []
        },
        {
            "id": "xvu",
            "web_id": 43210,
            "name": "Support Our Candidate List 2",
            "contact": {
                "company": "Support Our Candidate",
                "address1": "123 Main Street",
                "address2": "",
                "city": "Townsville",
                "state": "OH",
                "zip": "43358",
                "country": "US",
                "phone": ""
            },
            "permission_reminder": (
                "You are receiving this email because you signed up at an event, while being "
                "canvassed, or on our website."),
            "use_archive_bar": True,
            "campaign_defaults": {
                "from_name": "Our Candidate",
                "from_email": "our_candidate@example.com",
                "subject": "",
                "language": "en"
            },
            "notify_on_subscribe": "",
            "notify_on_unsubscribe": "",
            "date_created": "2018-09-15T22:15:21+00:00",
            "list_rating": 3,
            "email_type_option": False,
            "subscribe_url_short": "http://example.com/sample-subscribe_url_1",
            "subscribe_url_long": "https://mailchi.mp/zyx/sample-subscribe-url-1",
            "beamer_address": "us00-sample-1@inbound.mailchimp.com",
            "visibility": "pub",
            "double_optin": False,
            "has_welcome": True,
            "marketing_permissions": False,
            "modules": [],
            "stats": {
                "member_count": 73,
                "unsubscribe_count": 3,
                "cleaned_count": 7,
                "member_count_since_send": 1,
                "unsubscribe_count_since_send": 1,
                "cleaned_count_since_send": 0,
                "campaign_count": 13,
                "campaign_last_sent": "2020-01-03T14:38:11+00:00",
                "merge_field_count": 5,
                "avg_sub_rate": 0,
                "avg_unsub_rate": 1,
                "target_sub_rate": 3,
                "open_rate": 64.19236186394533,
                "click_rate": 3.746759370417411,
                "last_sub_date": "2020-01-01T00:19:46+00:00",
                "last_unsub_date": "2019-12-23T11:44:31+00:00"
            },
            "_links": []
        },
    ],
    "total_items": 1,
    "constraints": {
        "may_create": False,
        "max_instances": 1,
        "current_total_instances": 1
    },
    "_links": []
}

test_members = {
  "members": [
    {
      "id": "9eb69db8d0371811aa18803a1ae21584",
      "email_address": "member_1@example.com",
      "unique_email_id": "c82a25d939",
      "web_id": 24816326,
      "email_type": "html",
      "status": "subscribed",
      "merge_fields": {
        "FNAME": "Member",
        "LNAME": "One",
        "ADDRESS": {
          "addr1": "",
          "addr2": "",
          "city": "",
          "state": "",
          "zip": "",
          "country": "US"
        },
        "PHONE": "",
        "BIRTHDAY": ""
      },
      "stats": {
        "avg_open_rate": 0.3571,
        "avg_click_rate": 0
      },
      "ip_signup": "",
      "timestamp_signup": "",
      "ip_opt": "174.59.50.35",
      "timestamp_opt": "2019-03-25T22:55:44+00:00",
      "member_rating": 4,
      "last_changed": "2019-03-25T22:55:44+00:00",
      "language": "en",
      "vip": False,
      "email_client": "Gmail",
      "location": {
        "latitude": 40.0293,
        "longitude": -76.2656,
        "gmtoff": 0,
        "dstoff": 0,
        "country_code": "US",
        "timezone": "717/223"
      },
      "source": "Unknown",
      "tags_count": 0,
      "tags": [],
      "list_id": "67fdf4b1f4",
      "_links": []
    },
    {
      "id": "4f315641dbad7b74acc0f4a5d3741ac6",
      "email_address": "member_2@example.com",
      "unique_email_id": "8d308d69d3",
      "web_id": 12233445,
      "email_type": "html",
      "status": "subscribed",
      "merge_fields": {
        "FNAME": "Member",
        "LNAME": "Two",
        "ADDRESS": "",
        "PHONE": "",
        "BIRTHDAY": ""
      },
      "stats": {
        "avg_open_rate": 0.5,
        "avg_click_rate": 0
      },
      "ip_signup": "",
      "timestamp_signup": "",
      "ip_opt": "174.59.50.35",
      "timestamp_opt": "2019-03-25T23:04:46+00:00",
      "member_rating": 4,
      "last_changed": "2019-03-25T23:04:46+00:00",
      "language": "",
      "vip": False,
      "email_client": "iPhone",
      "location": {
        "latitude": 40.0459,
        "longitude": -76.3542,
        "gmtoff": 0,
        "dstoff": 0,
        "country_code": "US",
        "timezone": "717/223"
      },
      "source": "Import",
      "tags_count": 2,
      "tags": [
        {
          "id": 17493,
          "name": "canvass"
        },
        {
          "id": 17497,
          "name": "canvass-03-17-2019"
        }
      ],
      "list_id": "67fdf4b1f4",
      "_links": []
    }]}

test_unsubscribes = {
  "unsubscribes": [
    {
      "email_id": "e542e5cd7b414e5ff8409ff57cf154be",
      "email_address": "unsubscribe_1@exmaple.com",
      "merge_fields": {
        "FNAME": "Unsubscriber",
        "LNAME": "One",
        "ADDRESS": "",
        "PHONE": "5558754307",
        "BIRTHDAY": ""
      },
      "vip": False,
      "timestamp": "2019-12-09T21:18:06+00:00",
      "reason": "None given",
      "campaign_id": "abc",
      "list_id": "zyx",
      "list_is_active": True,
      "_links": []
    }
  ],
  "campaign_id": "abc",
  "total_items": 1,
  "_links": []
}



ENV_PARAMETERS = {
    'BLOOMERANG_API_KEY': 'env_api_key',
    'BLOOMERANG_CLIENT_ID': 'env_client_id',
    'BLOOMERANG_CLIENT_SECRET': 'env_client_secret'
}

ID = 123

TEST_DELETE = {
    "Id": 0,
    "Type": "string",
    "Deleted": 'true'
}

TEST_CREATE_CONSTITUENT = {
    "Type": "Individual",
    "Status": "Active",
    "FirstName": "string",
    "LastName": "string",
    "MiddleName": "string",
    "Prefix": "string",
    "Suffix": "string",
    "FullName": "string",
    "InformalName": "string",
    "FormalName": "string",
    "EnvelopeName": "string",
    "RecognitionName": "string",
    "JobTitle": "string",
    "Employer": "string",
    "Website": "string",
    "FacebookId": "string",
    "TwitterId": "string",
    "LinkedInId": "string",
    "Gender": "Male",
    "Birthdate": "2020-08-24",
    "ProfilePictureType": "None",
    "PrimaryEmail": {
        "Id": 0,
        "AccountId": 0,
        "Type": "Home",
        "Value": "user@example.com",
        "IsPrimary": 'true',
        "IsBad": 'true'
    },
    "PrimaryPhone": {
        "Id": 0,
        "AccountId": 0,
        "Type": "Home",
        "Extension": "string",
        "Number": "string",
        "IsPrimary": 'true'
    },
    "PrimaryAddress": {
        "Id": 0,
        "AccountId": 0,
        "Type": "Home",
        "Street": "string",
        "City": "string",
        "State": "string",
        "PostalCode": "string",
        "Country": "string",
        "IsPrimary": 'true',
        "IsBad": 'true'
    }
}

TEST_GET_CONSTITUENT = {
    "Id": 0,
    "AccountNumber": 0,
    "IsInHousehold": 'true',
    "IsHeadOfHousehold": 'true',
    "IsFavorite": 'true',
    "FullCustomProfileImageId": 0,
    "FullCustomProfileImageUrl": "string",
    "CroppedCustomProfileImageId": 0,
    "CroppedCustomProfileImageUrl": "string",
    "Type": "Individual",
    "Status": "Active",
    "FirstName": "string",
    "LastName": "string",
    "MiddleName": "string",
    "Prefix": "string",
    "Suffix": "string",
    "FullName": "string",
    "InformalName": "string",
    "FormalName": "string",
    "EnvelopeName": "string",
    "RecognitionName": "string",
    "JobTitle": "string",
    "Employer": "string",
    "Website": "string",
    "FacebookId": "string",
    "TwitterId": "string",
    "LinkedInId": "string",
    "Gender": "Male",
    "Birthdate": "2020-09-08",
    "ProfilePictureType": "None",
    "PrimaryEmail": {
        "Id": 0,
        "AccountId": 0,
        "Type": "Home",
        "Value": "user@example.com",
        "IsPrimary": 'true',
        "IsBad": 'true'
    },
    "PrimaryPhone": {
        "Id": 0,
        "AccountId": 0,
        "Type": "Home",
        "Extension": "string",
        "Number": "string",
        "IsPrimary": 'true'
    },
    "HouseholdId": 0,
    "PreferredCommunicationChannel": "Email",
    "CommunicationRestrictions": [
        "DoNotCall"
    ],
    "CommunicationRestrictionsUpdateReason": "string",
    "EmailInterestType": "All",
    "CustomEmailInterests": [
        {
            "Id": 0,
            "Name": "string"
        }
    ],
    "EmailInterestsUpdateReason": "string",
    "PrimaryAddress": {
        "Id": 0,
        "AccountId": 0,
        "Type": "Home",
        "Street": "string",
        "City": "string",
        "State": "string",
        "PostalCode": "string",
        "Country": "string",
        "IsPrimary": 'true',
        "IsBad": 'true',
        "StateAbbreviation": "string",
        "CountryCode": "string"
    },
    "EngagementScore": "Low",
    "DonorSearchInfo": {
        "Id": 0,
        "GenerosityScore": "Low",
        "AnnualFundLikelihood": "Low",
        "MajorGiftLikelihood": "Low",
        "PlannedGiftLikelihood": "Low",
        "Quality": "Low",
        "LargestGiftMin": 0,
        "LargestGiftMax": 0,
        "WealthAskMin": 0,
        "WealthAskMax": 0,
        "BusinessExecutive": 'true',
        "NamesScreened": "string",
        "DateTimeScreenedUtc": "string"
    },
    "AddressIds": [
        0
    ],
    "EmailIds": [
        0
    ],
    "PhoneIds": [
        0
    ],
    "CustomValues": [
        {
            "FieldId": 0,
            "Value": {
                "Id": 0,
                "Value": "string"
            }
        },
        {
            "FieldId": 0,
            "Values": [
                {
                    "Id": 0,
                    "Value": "string"
                }
            ]
        }
    ],
    "AuditTrail": {
        "CreatedDate": "2020-09-08T16:06:59.945Z",
        "CreatedName": "string",
        "LastModifiedDate": "2020-09-08T16:06:59.945Z",
        "LastModifiedName": "string"
    }
}

TEST_GET_CONSTITUENTS = {
    "Total": 0,
    "TotalFiltered": 0,
    "Start": 0,
    "ResultCount": 0,
    "Results": [
        {
            "Id": 0,
            "AccountNumber": 0,
            "IsInHousehold": 'true',
            "IsHeadOfHousehold": 'true',
            "IsFavorite": 'true',
            "FullCustomProfileImageId": 0,
            "FullCustomProfileImageUrl": "string",
            "CroppedCustomProfileImageId": 0,
            "CroppedCustomProfileImageUrl": "string",
            "Type": "Individual",
            "Status": "Active",
            "FirstName": "string",
            "LastName": "string",
            "MiddleName": "string",
            "Prefix": "string",
            "Suffix": "string",
            "FullName": "string",
            "InformalName": "string",
            "FormalName": "string",
            "EnvelopeName": "string",
            "RecognitionName": "string",
            "JobTitle": "string",
            "Employer": "string",
            "Website": "string",
            "FacebookId": "string",
            "TwitterId": "string",
            "LinkedInId": "string",
            "Gender": "Male",
            "Birthdate": "2020-09-05",
            "ProfilePictureType": "None",
            "PrimaryEmail": {
                "Id": 0,
                "AccountId": 0,
                "Type": "Home",
                "Value": "user@example.com",
                "IsPrimary": 'true',
                "IsBad": 'true'
            },
            "PrimaryPhone": {
                "Id": 0,
                "AccountId": 0,
                "Type": "Home",
                "Extension": "string",
                "Number": "string",
                "IsPrimary": 'true'
            },
            "HouseholdId": 0,
            "PreferredCommunicationChannel": "Email",
            "CommunicationRestrictions": [
                "DoNotCall"
            ],
            "CommunicationRestrictionsUpdateReason": "string",
            "EmailInterestType": "All",
            "CustomEmailInterests": [
                {
                    "Id": 0,
                    "Name": "string"
                }
            ],
            "EmailInterestsUpdateReason": "string",
            "EngagementScore": "Low",
            "DonorSearchInfo": {
                "Id": 0,
                "GenerosityScore": "Low",
                "AnnualFundLikelihood": "Low",
                "MajorGiftLikelihood": "Low",
                "PlannedGiftLikelihood": "Low",
                "Quality": "Low",
                "LargestGiftMin": 0,
                "LargestGiftMax": 0,
                "WealthAskMin": 0,
                "WealthAskMax": 0,
                "BusinessExecutive": 'true',
                "NamesScreened": "string",
                "DateTimeScreenedUtc": "string"
            },
            "AddressIds": [
                0
            ],
            "EmailIds": [
                0
            ],
            "PhoneIds": [
                0
            ],
            "CustomValues": [
                {
                    "FieldId": 0,
                    "Value": {
                        "Id": 0,
                        "Value": "string"
                    }
                },
                {
                    "FieldId": 0,
                    "Values": [
                        {
                            "Id": 0,
                            "Value": "string"
                        }
                    ]
                }
            ],
            "AuditTrail": {
                "CreatedDate": "2020-09-05T01:40:43.035Z",
                "CreatedName": "string",
                "LastModifiedDate": "2020-09-05T01:40:43.035Z",
                "LastModifiedName": "string"
            }
        }
    ]
}

TEST_CREATE_TRANSACTION = {
    "AccountId": 0,
    "Date": "2020-09-08",
    "Amount": 0,
    "Method": "None",
    "CheckNumber": "string",
    "CheckDate": "2020-09-08",
    "CreditCardType": "Visa",
    "CreditCardLastFourDigits": "string",
    "CreditCardExpMonth": 0,
    "CreditCardExpYear": 0,
    "EftAccountType": "Checking",
    "EftLastFourDigits": "string",
    "EftRoutingNumber": "string",
    "InKindDescription": "string",
    "InKindType": "Goods",
    "InKindMarketValue": 0,
    "WalletItemId": 0,
    "IntegrationUrl": "string",
    "Designations": [
        {
            "Amount": 0,
            "Note": "string",
            "AcknowledgementStatus": 'true',
            "AcknowledgementInteractionIds": [
                0
            ],
            "Type": "Donation",
            "NonDeductibleAmount": 0,
            "FundId": 0,
            "QuickbooksAccountId": 0,
            "CampaignId": 0,
            "AppealId": 0,
            "TributeId": 0,
            "SoftCredits": [
                'null'
            ],
            "CustomValues": [
                {
                    "FieldId": 0,
                    "Value": "string"
                },
                {
                    "FieldId": 0,
                    "ValueId": 0
                },
                {
                    "FieldId": 0,
                    "ValueIds": [
                        0
                    ]
                }
            ],
            "Attachments": [
                'null',
                'null'
            ]
        },
        {
            "Amount": 0,
            "Note": "string",
            "AcknowledgementStatus": 'true',
            "AcknowledgementInteractionIds": [
                0
            ],
            "Type": "PledgePayment",
            "PledgeId": 0,
            "CustomValues": [
                {
                    "FieldId": 0,
                    "Value": "string"
                },
                {
                    "FieldId": 0,
                    "ValueId": 0
                },
                {
                    "FieldId": 0,
                    "ValueIds": [
                        0
                    ]
                }
            ],
            "Attachments": [
                'null',
                'null'
            ]
        },
        {
            "RecurringDonationEndDate": "2020-09-08",
            "Amount": 0,
            "Note": "string",
            "AcknowledgementStatus": 'true',
            "AcknowledgementInteractionIds": [
                0
            ],
            "RecurringDonationFrequency": "Weekly",
            "RecurringDonationDay1": 0,
            "RecurringDonationDay2": 0,
            "RecurringDonationStartDate": "2020-09-08",
            "Type": "RecurringDonation",
            "FundId": 0,
            "QuickbooksAccountId": 0,
            "CampaignId": 0,
            "AppealId": 0,
            "TributeId": 0,
            "SoftCredits": [
                'null'
            ],
            "CustomValues": [
                {
                    "FieldId": 0,
                    "Value": "string"
                },
                {
                    "FieldId": 0,
                    "ValueId": 0
                },
                {
                    "FieldId": 0,
                    "ValueIds": [
                        0
                    ]
                }
            ],
            "Attachments": [
                'null',
                'null'
            ]
        },
        {
            "Amount": 0,
            "Note": "string",
            "AcknowledgementStatus": 'true',
            "AcknowledgementInteractionIds": [
                0
            ],
            "Type": "RecurringDonationPayment",
            "RecurringDonationId": 0,
            "FundId": 0,
            "QuickbooksAccountId": 0,
            "CampaignId": 0,
            "AppealId": 0,
            "IsExtraPayment": 'true',
            "CustomValues": [
                {
                    "FieldId": 0,
                    "Value": "string"
                },
                {
                    "FieldId": 0,
                    "ValueId": 0
                },
                {
                    "FieldId": 0,
                    "ValueIds": [
                        0
                    ]
                }
            ],
            "Attachments": [
                'null',
                'null'
            ]
        }
    ],
    "Attachments": [
        {
            "Guid": "string",
            "Name": "string",
            "Extension": "string",
            "Url": "string"
        },
        {
            "Id": 0,
            "Name": "string",
            "Extension": "string",
            "Url": "string"
        }
    ]
}

TEST_GET_TRANSACTION = {
    "Id": 0,
    "TransactionNumber": 0,
    "NonDeductibleAmount": 0,
    "AccountId": 0,
    "Date": "2020-09-08",
    "Amount": 0,
    "Method": "None",
    "CheckNumber": "string",
    "CheckDate": "2020-09-08",
    "CreditCardType": "Visa",
    "CreditCardLastFourDigits": "string",
    "CreditCardExpMonth": 0,
    "CreditCardExpYear": 0,
    "EftAccountType": "Checking",
    "EftLastFourDigits": "string",
    "EftRoutingNumber": "string",
    "InKindDescription": "string",
    "InKindType": "Goods",
    "InKindMarketValue": 0,
    "IntegrationUrl": "string",
    "Designations": [
        {
            "Id": 0,
            "DesignationNumber": 0,
            "TransactionId": 0,
            "Amount": 0,
            "NonDeductibleAmount": 0,
            "Note": "string",
            "AcknowledgementStatus": 'true',
            "AcknowledgementInteractionIds": [
                0
            ],
            "Fund": {
                "Id": 0,
                "Name": "string"
            },
            "QuickbooksAccount": {
                "Id": 0,
                "Name": "string"
            },
            "Campaign": {
                "Id": 0,
                "Name": "string"
            },
            "Appeal": {
                "Id": 0,
                "Name": "string"
            },
            "Tribute": {
                "Id": 0,
                "Name": "string"
            },
            "TributeType": "InHonorOf",
            "SoftCreditIds": [
                0
            ],
            "AttachmentIds": [
                0
            ],
            "CustomValues": [
                {
                    "FieldId": 0,
                    "Value": {
                        "Id": 0,
                        "Value": "string"
                    }
                },
                {
                    "FieldId": 0,
                    "Values": [
                        {
                            "Id": 0,
                            "Value": "string"
                        }
                    ]
                }
            ],
            "AuditTrail": {
                "CreatedDate": "2020-09-08T14:15:24.293Z",
                "CreatedName": "string",
                "LastModifiedDate": "2020-09-08T14:15:24.293Z",
                "LastModifiedName": "string"
            },
            "Type": "Donation"
        },
        {
            "Id": 0,
            "DesignationNumber": 0,
            "TransactionId": 0,
            "Amount": 0,
            "NonDeductibleAmount": 0,
            "Note": "string",
            "AcknowledgementStatus": 'true',
            "AcknowledgementInteractionIds": [
                0
            ],
            "Fund": {
                "Id": 0,
                "Name": "string"
            },
            "QuickbooksAccount": {
                "Id": 0,
                "Name": "string"
            },
            "Campaign": {
                "Id": 0,
                "Name": "string"
            },
            "Appeal": {
                "Id": 0,
                "Name": "string"
            },
            "Tribute": {
                "Id": 0,
                "Name": "string"
            },
            "TributeType": "InHonorOf",
            "SoftCreditIds": [
                0
            ],
            "AttachmentIds": [
                0
            ],
            "CustomValues": [
                {
                    "FieldId": 0,
                    "Value": {
                        "Id": 0,
                        "Value": "string"
                    }
                },
                {
                    "FieldId": 0,
                    "Values": [
                        {
                            "Id": 0,
                            "Value": "string"
                        }
                    ]
                }
            ],
            "AuditTrail": {
                "CreatedDate": "2020-09-08T14:15:24.293Z",
                "CreatedName": "string",
                "LastModifiedDate": "2020-09-08T14:15:24.293Z",
                "LastModifiedName": "string"
            },
            "Type": "Pledge",
            "PledgePaymentIds": [
                0
            ],
            "PledgeInstallments": [
                {
                    "Id": 0,
                    "PledgeId": 0,
                    "Date": "2020-09-08",
                    "Amount": 0
                }
            ],
            "PledgeBalance": 0,
            "PledgeStatus": "InGoodStanding",
            "PledgeAmountInArrears": 0,
            "PledgeNextInstallmentDate": "2020-09-08"
        },
        {
            "Id": 0,
            "DesignationNumber": 0,
            "TransactionId": 0,
            "Amount": 0,
            "NonDeductibleAmount": 0,
            "Note": "string",
            "AcknowledgementStatus": 'true',
            "AcknowledgementInteractionIds": [
                0
            ],
            "Fund": {
                "Id": 0,
                "Name": "string"
            },
            "QuickbooksAccount": {
                "Id": 0,
                "Name": "string"
            },
            "Campaign": {
                "Id": 0,
                "Name": "string"
            },
            "Appeal": {
                "Id": 0,
                "Name": "string"
            },
            "Tribute": {
                "Id": 0,
                "Name": "string"
            },
            "TributeType": "InHonorOf",
            "SoftCreditIds": [
                0
            ],
            "AttachmentIds": [
                0
            ],
            "CustomValues": [
                {
                    "FieldId": 0,
                    "Value": {
                        "Id": 0,
                        "Value": "string"
                    }
                },
                {
                    "FieldId": 0,
                    "Values": [
                        {
                            "Id": 0,
                            "Value": "string"
                        }
                    ]
                }
            ],
            "AuditTrail": {
                "CreatedDate": "2020-09-08T14:15:24.293Z",
                "CreatedName": "string",
                "LastModifiedDate": "2020-09-08T14:15:24.293Z",
                "LastModifiedName": "string"
            },
            "Type": "PledgePayment",
            "PledgeId": 0
        },
        {
            "Id": 0,
            "DesignationNumber": 0,
            "TransactionId": 0,
            "Amount": 0,
            "NonDeductibleAmount": 0,
            "Note": "string",
            "AcknowledgementStatus": 'true',
            "AcknowledgementInteractionIds": [
                0
            ],
            "Fund": {
                "Id": 0,
                "Name": "string"
            },
            "QuickbooksAccount": {
                "Id": 0,
                "Name": "string"
            },
            "Campaign": {
                "Id": 0,
                "Name": "string"
            },
            "Appeal": {
                "Id": 0,
                "Name": "string"
            },
            "Tribute": {
                "Id": 0,
                "Name": "string"
            },
            "TributeType": "InHonorOf",
            "SoftCreditIds": [
                0
            ],
            "AttachmentIds": [
                0
            ],
            "CustomValues": [
                {
                    "FieldId": 0,
                    "Value": {
                        "Id": 0,
                        "Value": "string"
                    }
                },
                {
                    "FieldId": 0,
                    "Values": [
                        {
                            "Id": 0,
                            "Value": "string"
                        }
                    ]
                }
            ],
            "AuditTrail": {
                "CreatedDate": "2020-09-08T14:15:24.293Z",
                "CreatedName": "string",
                "LastModifiedDate": "2020-09-08T14:15:24.293Z",
                "LastModifiedName": "string"
            },
            "RecurringDonationEndDate": "2020-09-08",
            "RecurringDonationFrequency": "Weekly",
            "RecurringDonationDay1": 0,
            "RecurringDonationDay2": 0,
            "RecurringDonationStartDate": "2020-09-08",
            "Type": "RecurringDonation",
            "RecurringDonationPaymentIds": [
                0
            ],
            "RecurringDonationNextInstallmentDate": "2020-09-08",
            "RecurringDonationLastPaymentStatus": "AtRisk",
            "RecurringDonationStatus": "Active"
        },
        {
            "Id": 0,
            "DesignationNumber": 0,
            "TransactionId": 0,
            "Amount": 0,
            "NonDeductibleAmount": 0,
            "Note": "string",
            "AcknowledgementStatus": 'true',
            "AcknowledgementInteractionIds": [
                0
            ],
            "Fund": {
                "Id": 0,
                "Name": "string"
            },
            "QuickbooksAccount": {
                "Id": 0,
                "Name": "string"
            },
            "Campaign": {
                "Id": 0,
                "Name": "string"
            },
            "Appeal": {
                "Id": 0,
                "Name": "string"
            },
            "Tribute": {
                "Id": 0,
                "Name": "string"
            },
            "TributeType": "InHonorOf",
            "SoftCreditIds": [
                0
            ],
            "AttachmentIds": [
                0
            ],
            "CustomValues": [
                {
                    "FieldId": 0,
                    "Value": {
                        "Id": 0,
                        "Value": "string"
                    }
                },
                {
                    "FieldId": 0,
                    "Values": [
                        {
                            "Id": 0,
                            "Value": "string"
                        }
                    ]
                }
            ],
            "AuditTrail": {
                "CreatedDate": "2020-09-08T14:15:24.293Z",
                "CreatedName": "string",
                "LastModifiedDate": "2020-09-08T14:15:24.293Z",
                "LastModifiedName": "string"
            },
            "Type": "RecurringDonationPayment",
            "RecurringDonationId": 0
        }
    ],
    "AttachmentIds": [
        0
    ],
    "IsRefunded": 'true',
    "RefundIds": [
        0
    ],
    "AuditTrail": {
        "CreatedDate": "2020-09-08T14:15:24.293Z",
        "CreatedName": "string",
        "LastModifiedDate": "2020-09-08T14:15:24.293Z",
        "LastModifiedName": "string"
    }
}

TEST_GET_TRANSACTIONS = {
    "Total": 0,
    "TotalFiltered": 0,
    "Start": 0,
    "ResultCount": 0,
    "Results": [
        {
            "Id": 0,
            "TransactionNumber": 0,
            "NonDeductibleAmount": 0,
            "AccountId": 0,
            "Date": "2020-09-08",
            "Amount": 0,
            "Method": "None",
            "CheckNumber": "string",
            "CheckDate": "2020-09-08",
            "CreditCardType": "Visa",
            "CreditCardLastFourDigits": "string",
            "CreditCardExpMonth": 0,
            "CreditCardExpYear": 0,
            "EftAccountType": "Checking",
            "EftLastFourDigits": "string",
            "EftRoutingNumber": "string",
            "InKindDescription": "string",
            "InKindType": "Goods",
            "InKindMarketValue": 0,
            "IntegrationUrl": "string",
            "Designations": [
                'null',
                'null',
                'null',
                'null',
                'null'
            ],
            "AttachmentIds": [
                0
            ],
            "IsRefunded": 'true',
            "RefundIds": [
                0
            ],
            "AuditTrail": {
                "CreatedDate": "2020-09-08T16:11:30.821Z",
                "CreatedName": "string",
                "LastModifiedDate": "2020-09-08T16:11:30.821Z",
                "LastModifiedName": "string"
            }
        }
    ]
}

TEST_CREATE_INTERACTION = {
    "AccountId": 0,
    "Date": "2020-09-08",
    "Note": "string",
    "Channel": "Email",
    "Purpose": "Acknowledgement",
    "Subject": "string",
    "IsInbound": 'true',
    "CustomValues": [
        {
            "FieldId": 0,
            "Value": "string"
        },
        {
            "FieldId": 0,
            "ValueId": 0
        },
        {
            "FieldId": 0,
            "ValueIds": [
                0
            ]
        }
    ],
    "Attachments": [
        {
            "Guid": "string",
            "Name": "string",
            "Extension": "string",
            "Url": "string"
        },
        {
            "Id": 0,
            "Name": "string",
            "Extension": "string",
            "Url": "string"
        }
    ]
}

TEST_GET_INTERACTION = {
    "Id": 0,
    "Date": "2020-09-08",
    "Note": "string",
    "Channel": "Email",
    "Purpose": "Acknowledgement",
    "Subject": "string",
    "IsInbound": 'true',
    "AccountId": 0,
    "TweetId": "string",
    "IsBcc": 'true',
    "EmailAddress": "user@example.com",
    "AttachmentIds": [
        0
    ],
    "LetterAttachmentIds": [
        0
    ],
    "SurveyLapsedResponses": [
        "string"
    ],
    "SurveyEmailInteractionId": 0,
    "SurveyResponseInteractionId": 0,
    "CustomValues": [
        {
            "FieldId": 0,
            "Value": {
                "Id": 0,
                "Value": "string"
            }
        },
        {
            "FieldId": 0,
            "Values": [
                {
                    "Id": 0,
                    "Value": "string"
                }
            ]
        }
    ],
    "AuditTrail": {
        "CreatedDate": "2020-09-08T15:27:32.767Z",
        "CreatedName": "string",
        "LastModifiedDate": "2020-09-08T15:27:32.767Z",
        "LastModifiedName": "string"
    }
}

TEST_GET_INTERACTIONS = {
    "Total": 0,
    "TotalFiltered": 0,
    "Start": 0,
    "ResultCount": 0,
    "Results": [
        {
            "Id": 0,
            "Date": "2020-09-08",
            "Note": "string",
            "Channel": "Email",
            "Purpose": "Acknowledgement",
            "Subject": "string",
            "IsInbound": 'true',
            "AccountId": 0,
            "TweetId": "string",
            "IsBcc": 'true',
            "EmailAddress": "user@example.com",
            "AttachmentIds": [
                0
            ],
            "LetterAttachmentIds": [
                0
            ],
            "SurveyLapsedResponses": [
                "string"
            ],
            "SurveyEmailInteractionId": 0,
            "SurveyResponseInteractionId": 0,
            "CustomValues": [
                {
                    "FieldId": 0,
                    "Value": {
                        "Id": 0,
                        "Value": "string"
                    }
                },
                {
                    "FieldId": 0,
                    "Values": [
                        {
                            "Id": 0,
                            "Value": "string"
                        }
                    ]
                }
            ],
            "AuditTrail": {
                "CreatedDate": "2020-09-08T16:10:36.389Z",
                "CreatedName": "string",
                "LastModifiedDate": "2020-09-08T16:10:36.389Z",
                "LastModifiedName": "string"
            }
        }
    ]
}



import logging
from functools import partial, wraps

import petl
import requests
from github import Github as PyGithub
from github.GithubException import UnknownObjectException

from parsons.etl.table import Table
from parsons.utilities import check_env, files

logger = logging.getLogger(__name__)


def _wrap_method(decorator, method):
    def _wrapper(self, *args, **kwargs):
        bound_method = partial(method.__get__(self, type(self)))
        return decorator(bound_method)(*args, **kwargs)
    return _wrapper


def decorate_methods(decorator):
    # Based on Django's django.utils.decorators.method_decorator
    def decorate(cls):
        for method in dir(cls):
            # Don't decorate dunder methods
            if method.startswith("__"):
                continue
            cls_method = getattr(cls, method)
            if callable(cls_method):
                setattr(cls, method, _wrap_method(decorator, cls_method))
        return cls
    return decorate


def wrap_github_404(func):
    @wraps(func)
    def _wrapped_func(*args, **kwargs):
        try:
            return (func)(*args, **kwargs)
        except UnknownObjectException:
            raise ParsonsGitHubError(
                "Couldn't find the object you referenced, maybe you need to log in?"
            )
    return _wrapped_func


class ParsonsGitHubError(Exception):
    pass


@decorate_methods(wrap_github_404)
class GitHub(object):
    """Creates a GitHub class for accessing the GitHub API.

    Uses ``parsons.utilities.check_env`` to load credentials from environment variables if not
    supplied. Supports either a username and password or an access token for authentication. The
    client also supports unauthenticated access.

    Args:
        username: Optional[str]
            Username of account to use for credentials. Can be set with ``GITHUB_USERNAME``
            environment variable.
        password: Optional[str]
            Password of account to use for credentials. Can be set with ``GITHUB_PASSWORD``
            environment variable.
        access_token: Optional[str]
            Access token to use for credentials. Can be set with ``GITHUB_ACCESS_TOKEN`` environment
            variable.
    """

    def __init__(self, username=None, password=None, access_token=None):

        self.username = check_env.check('GITHUB_USERNAME', username, optional=True)
        self.password = check_env.check('GITHUB_PASSWORD', password, optional=True)
        self.access_token = check_env.check('GITHUB_ACCESS_TOKEN', access_token, optional=True)

        if self.username and self.password:
            self.client = PyGithub(self.username, self.password)
        elif self.access_token:
            self.client = PyGithub(self.access_token)
        else:
            self.client = PyGithub()

    def _as_table(self, paginated_list, page=None, page_size=100):
        """Converts a paginated list into a Parsons ``Table``. Uses the ``_rawData`` property of
        each item instead of calling ``raw_data`` to avoid making a separate request for each item
        in a page for types that PyGithub doesn't consider complete.

        Args:
            paginated_list: ``pygithub.PaginatedList.PaginatedList``
                PyGithub paginated list
            page: Optional[int]
                Page number to load. Defaults to None. If not specified, all results are returned.
            page_size: int
                Page size. Defaults to 100. Ignored if ``page`` is not set.

        Returns:
            ``Table``
                Table object created from the raw data of the list
        """

        if page is not None:
            page_start = (page - 1) * page_size
            page_end = page_start + page_size
            list_pages = paginated_list[page_start:page_end]
        else:
            list_pages = paginated_list

        return Table([list_item._rawData for list_item in list_pages])

    def get_user(self, username):
        """Loads a GitHub user by username

        Args:
            username: str
                Username of user to load

        Returns:
            dict
                User information
        """

        return self.client.get_user(username).raw_data

    def get_organization(self, organization_name):
        """Loads a GitHub organization by name

        Args:
            organization_name: str
                Name of organization to load

        Returns:
            dict
                Organization information
        """

        return self.client.get_organization(organization_name).raw_data

    def get_repo(self, repo_name):
        """Loads a GitHub repo by name

        Args:
            repo_name: str
                Full repo name (account/name)

        Returns:
            dict
                Repo information
        """

        return self.client.get_repo(repo_name).raw_data

    def list_user_repos(self, username, page=None, page_size=100):
        """List user repos with pagination, returning a ``Table``

        Args:
            username: str
                GitHub username
            page: Optional[int]
                Page number. All results are returned if not set.
            page_size: int
                Page size. Defaults to 100.

        Returns:
            ``Table``
                Table with page of user repos
        """

        logger.info(f'Listing page {page} of repos for user {username}')

        return self._as_table(
            self.client.get_user(username).get_repos(), page=page, page_size=page_size
        )

    def list_organization_repos(self, organization_name, page=None, page_size=100):
        """List organization repos with pagination, returning a ``Table``

        Args:
            organization_name: str
                GitHub organization name
            page: Optional[int]
                Page number. All results are returned if not set.
            page_size: int
                Page size. Defaults to 100.

        Returns:
            ``Table``
                Table with page of organization repos
        """

        logger.info(f'Listing page {page} of repos for organization {organization_name}')

        return self._as_table(
            self.client.get_organization(organization_name).get_repos(),
            page=page,
            page_size=page_size,
        )

    def get_issue(self, repo_name, issue_number):
        """Loads a GitHub issue

        Args:
            repo_name: str
                Full repo name (account/name)
            issue_number: int
                Number of issue to load

        Returns:
            dict
                Issue information
        """

        return self.client.get_repo(repo_name).get_issue(number=issue_number).raw_data

    def list_repo_issues(self, repo_name, state="open", assignee=None, creator=None, mentioned=None,
                         labels=[], sort="created", direction="desc", since=None, page=None,
                         page_size=100):
        """List issues for a given repo

        Args:
            repo_name: str
                Full repo name (account/name)
            state: str
                State of issues to return. One of "open", "closed", "all". Defaults to "open".
            assignee: Optional[str]
                Name of assigned user, "none", or "*".
            creator: Optional[str]
                Name of user that created the issue.
            mentioned: Optional[str]
                Name of user mentioned in the issue.
            labels: list[str]
                List of label names. Defaults to []
            sort: str
                What to sort results by. One of "created", "updated", "comments". Defaults to
                "created".
            direction: str
                Direction to sort. One of "asc", "desc". Defaults to "desc".
            since: Optional[Union[datetime.datetime, datetime.date]]
                Timestamp to pull issues since. Defaults to None.
            page: Optional[int]
                Page number. All results are returned if not set.
            page_size: int
                Page size. Defaults to 100.

        Returns:
            ``Table``
                Table with page of repo issues
        """

        logger.info(f'Listing page {page} of issues for repo {repo_name}')

        kwargs_dict = {"state": state, "sort": sort, "direction": direction}
        if assignee:
            kwargs_dict["assignee"] = assignee
        if creator:
            kwargs_dict["creator"] = creator
        if mentioned:
            kwargs_dict["mentioned"] = mentioned
        if len(labels) > 0:
            kwargs_dict["labels"] = ",".join(labels)
        if since:
            kwargs_dict["since"] = f'{since.isoformat()[:19]}Z'

        return self._as_table(
            self.client.get_repo(repo_name).get_issues(**kwargs_dict),
            page=page,
            page_size=page_size,
        )

    def get_pull_request(self, repo_name, pull_request_number):
        """Loads a GitHub pull request

        Args:
            repo_name: str
                Full repo name (account/name)
            pull_request_number: int
                Pull request number

        Returns:
            dict
                Pull request information
        """

        return self.client.get_repo(repo_name).get_pull(pull_request_number).raw_data

    def list_repo_pull_requests(self, repo_name, state="open", base=None, sort="created",
                                direction="desc", page=None, page_size=100):
        """Lists pull requests for a given repo

        Args:
            repo_name: str
                Full repo name (account/name)
            state: str
                One of "open, "closed", "all". Defaults to "open".
            base: Optional[str]
                Base branch to filter pull requests by.
            sort: str
                How to sort pull requests. One of "created", "updated", "popularity". Defaults to
                "created".
            direction: str
                Direction to sort by. Defaults to "desc".
            page: Optional[int]
                Page number. All results are returned if not set.
            page_size: int
                Page size. Defaults to 100.

        Returns:
            ``Table``
                Table with page of repo pull requests
        """

        logger.info(f'Listing page {page} of pull requests for repo {repo_name}')

        kwargs_dict = {"state": state, "sort": sort, "direction": direction}
        if base:
            kwargs_dict["base"] = base

        self._as_table(
            self.client.get_repo(repo_name).get_pulls(**kwargs_dict), page=page, page_size=page_size
        )

    def list_repo_contributors(self, repo_name, page=None, page_size=100):
        """Lists contributors for a given repo

        Args:
            repo_name: str
                Full repo name (account/name)
            page: Optional[int]
                Page number. All results are returned if not set.
            page_size: int
                Page size. Defaults to 100.

        Returns:
            ``Table``
                Table with page of repo contributors
        """

        logger.info(f'Listing page {page} of contributors for repo {repo_name}')

        return self._as_table(
            self.client.get_repo(repo_name).get_contributors(), page=page, page_size=page_size
        )

    def download_file(self, repo_name, path, branch=None, local_path=None):
        """Download a file from a repo by path and branch. Defaults to the repo's default branch if
        branch is not supplied.

        Uses the download_url directly rather than the API because the API only supports contents up
        to 1MB from a repo directly, and the process for downloading larger files through the API is
        much more involved.

        Because download_url does not go through the API, it does not support username / password
        authentication, and requires a token to authenticate.

        Args:
            repo_name: str
                Full repo name (account/name)
            path: str
                Path from the repo base directory
            branch: Optional[str]
                Branch to download file from. Defaults to repo default branch
            local_path: Optional[str]
                Local file path to download file to. Will create a temp file if not supplied.

        Returns:
            str
                File path of downloaded file
        """

        if not local_path:
            local_path = files.create_temp_file_for_path(path)

        repo = self.client.get_repo(repo_name)
        if branch is None:
            branch = repo.default_branch

        logger.info(f'Downloading {path} from {repo_name}, branch {branch} to {local_path}')

        headers = None
        if self.access_token:
            headers = {
                'Authorization': f'token {self.access_token}',
            }

        res = requests.get(f'https://raw.githubusercontent.com/{repo_name}/{branch}/{path}',
                           headers=headers)

        if res.status_code == 404:
            raise UnknownObjectException(status=404, data=res.content)
        elif res.status_code != 200:
            raise ParsonsGitHubError(
                f'Error downloading {path} from repo {repo_name}: {res.content}')

        with open(local_path, 'wb') as f:
            f.write(res.content)

        logger.info(f'Downloaded {path} to {local_path}')

        return local_path

    def download_table(self, repo_name, path, branch=None, local_path=None, delimiter=','):
        """Download a CSV file from a repo by path and branch as a Parsons Table.

        Args:
            repo_name: str
                Full repo name (account/name)
            path: str
                Path from the repo base directory
            branch: Optional[str]
                Branch to download file from. Defaults to repo default branch
            local_path: Optional[str]
                Local file path to download file to. Will create a temp file if not supplied.
            delimiter: Optional[str]
                The CSV delimiter to use to parse the data. Defaults to ','

        Returns:
            Parsons Table
                See :ref:`parsons-table` for output options.
        """
        downloaded_file = self.download_file(repo_name, path, branch, local_path)

        return Table(petl.fromcsv(downloaded_file, delimiter=delimiter))



from dateutil.parser import parse


class Flags:
    """A class to access the Flags PDI API endpoint."""

    def __init__(self):
        flags_endpoint = "/flags"
        self.url_flags = self.base_url + flags_endpoint

        super().__init__()

    def get_flags(self, start_date, end_date, limit=None):
        """Get a list of flags.

        `Args:`
            start_date: str
                A start date formatted like yyyy-MM-dd.
            end_date: str
                An end date formatted like yyyy-MM-dd.
            limit: int
                Specify limit to return.

        `Returns:`
            parsons.Table
                A Parsons table of all the data.
        """
        try:
            start_date = parse(start_date).date().isoformat()
            end_date = parse(end_date).date().isoformat()
        except ValueError:
            raise ValueError("Invalid date format.")

        params = {
            "startDate": start_date,
            "endDate": end_date,
        }

        return self._request(self.url_flags, args=params, limit=limit)



"""NGPVAN Bulk Import Endpoints"""
from parsons.etl.table import Table
from parsons.utilities import cloud_storage

import logging
import uuid
import csv

logger = logging.getLogger(__name__)


class BulkImport(object):

    def __init__(self):

        pass

    def get_bulk_import_resources(self):
        """
        Get bulk import resources that available to the user. These define
        the types of bulk imports that you can run. These might include
        ``Contacts``, ``ActivistCodes``, ``ContactsActivistCodes`` and others.

        `Returns:`
            list
                A list of resources.
        """

        r = self.connection.get_request(f'bulkImportJobs/resources')
        logger.info(f'Found {len(r)} bulk import resources.')
        return r

    def get_bulk_import_job(self, job_id):
        """
        Get a bulk import job status.

        `Args:`
            job_id : int
                The bulk import job id.
        `Returns:`
            dict
                The bulk import job
        """

        r = self.connection.get_request(f'bulkImportJobs/{job_id}')
        logger.info(f'Found bulk import job {job_id}.')
        return r

    def get_bulk_import_mapping_types(self):
        """
        Get bulk import mapping types.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self.connection.get_request('bulkImportMappingTypes'))
        logger.info(f'Found {tbl.num_rows} bulk import mapping types.')
        return tbl

    def get_bulk_import_mapping_type(self, type_name):
        """
        Get a single bulk import mapping type.

        `Args:`
            type_name: str
        `Returns`:
            dict
                A mapping type json
        """

        r = self.connection.get_request(f'bulkImportMappingTypes/{type_name}')
        logger.info(f'Found {type_name} bulk import mapping type.')
        return r

    def post_bulk_import(self, tbl, url_type, resource_type, mapping_types,
                         description, **url_kwargs):
        # Internal method to post bulk imports.

        # Move to cloud storage
        file_name = str(uuid.uuid1())
        url = cloud_storage.post_file(tbl,
                                      url_type,
                                      file_path=file_name + '.zip',
                                      quoting=csv.QUOTE_ALL,
                                      **url_kwargs)
        logger.info(f'Table uploaded to {url_type}.')

        # Generate request json
        json = {"description": description,
                "file": {
                    "columnDelimiter": 'csv',
                    "columns": [{'name': c} for c in tbl.columns],
                    "fileName": file_name + '.csv',
                    "hasHeader": "True",
                    "hasQuotes": "True",
                    "sourceUrl": url},
                "actions": [{"resultFileSizeKbLimit": 5000,
                             "resourceType": resource_type,
                             "actionType": "loadMappedFile",
                             "mappingTypes": mapping_types}]
                }

        r = self.connection.post_request('bulkImportJobs', json=json)
        logger.info(f"Bulk upload {r['jobId']} created.")
        return r['jobId']

    def bulk_apply_activist_codes(self, tbl, url_type, **url_kwargs):
        """
        Bulk apply activist codes.

        The table may include the following columns. The first column
        must be ``vanid``.

        .. list-table::
            :widths: 25 25 50
            :header-rows: 1

            * - Column Name
              - Required
              - Description
            * - ``vanid``
              - Yes
              - A valid VANID primary key
            * - ``activistcodeid``
              - Yes
              - A valid activist code id
            * - ``datecanvassed``
              - No
              - An ISO formatted date
            * - ``contacttypeid``
              - No
              - The method of contact.

        `Args:`
            table: Parsons table
                A Parsons table.
            url_type: str
                The cloud file storage to use to post the file. Currently only ``S3``.
            **url_kwargs: kwargs
                Arguments to configure your cloud storage url type.
                    * S3 requires ``bucket`` argument and, if not stored as env variables
                      ``aws_access_key`` and ``aws_secret_access_key``.
        `Returns:`
            int
                The bulk import job id
        """

        return self.post_bulk_import(tbl,
                                     url_type,
                                     'ContactsActivistCodes',
                                     [{"name": "ActivistCode"}],
                                     'Activist Code Upload',
                                     **url_kwargs)



from datetime import datetime, timedelta
import re

from parsons.etl.table import Table
from parsons.utilities import check_env
from parsons.utilities.api_connector import APIConnector


class Shopify(object):
    """
    Instantiate the Shopify class

    `Args:`
        subdomain: str
            The Shopify subdomain (e.g. ``myorg`` for myorg.myshopify.com) Not required if
            ``SHOPIFY_SUBDOMAIN`` env variable set.
        password: str
            The Shopify account password. Not required if ``SHOPIFY_PASSWORD`` env
            variable set.
        api_key: str
            The Shopify account API key. Not required if ``SHOPIFY_API_KEY`` env variable
            set.
        api_version: str
            The Shopify API version. Not required if ``SHOPIFY_API_VERSION`` env variable
            set.
    `Returns:`
        Shopify Class
    """
    def __init__(self, subdomain=None, password=None, api_key=None, api_version=None):
        self.subdomain = check_env.check('SHOPIFY_SUBDOMAIN', subdomain)
        self.password = check_env.check('SHOPIFY_PASSWORD', password)
        self.api_key = check_env.check('SHOPIFY_API_KEY', api_key)
        self.api_version = check_env.check('SHOPIFY_API_VERSION', api_version)
        self.client = APIConnector('https://%s.myshopify.com/admin/api/%s/' % (
            self.subdomain,
            self.api_version
        ), auth=(self.api_key, self.password))

    def get_count(self, query_date=None, since_id=None, table_name=None):
        """
        Get the count of rows in a table.

        `Args:`
            query_date: str
                Filter query by a date that rows were created. This filter is ignored if value
                is None.
            since_id: str
                Filter query by a minimum ID. This filter is ignored if value is None.
            table_name: str
                The name of the Shopify table to query.
        `Returns:`
            int
        """
        return self.client.request(self.get_query_url(query_date, since_id, table_name),
                                   'GET').json().get("count", 0)

    def get_orders(self, query_date=None, since_id=None, completed=True):
        """
        Get Shopify orders.

        `Args:`
            query_date: str
                Filter query by a date that rows were created. Format: yyyy-mm-dd. This filter
                is ignored if value is None.
            since_id: str
                Filter query by a minimum ID. This filter is ignored if value is None.
            completed: bool
                True if only getting completed orders, False otherwise.
        `Returns:`
            Table Class
        """
        orders = []

        def _append_orders(url):
            nonlocal orders

            if completed:
                url += '&financial_status=paid'

            res = self.client.request(url, 'GET')

            cur_orders = res.json().get("orders", [])

            # Flatten orders to non-complex types
            for order in cur_orders:
                keys_to_add = {}
                keys_to_delete = []

                for key1 in order:
                    if isinstance(order[key1], dict):
                        for key2 in order[key1]:
                            keys_to_add[key1 + '_' + key2] = order[key1][key2]
                        keys_to_delete.append(key1)
                    elif key1 == 'note_attributes':
                        for note in order[key1]:
                            keys_to_add[key1 + '_' + note['name']] = note['value']

                order.update(keys_to_add)
                for key in keys_to_delete:
                    del order[key]

            orders += cur_orders

            return res

        res = _append_orders(self.get_query_url(query_date, since_id, "orders", False))

        # Get next page
        while res.headers.get("Link"):
            link = re.split('; |, ', res.headers.get("Link"))
            if len(link) and link[len(link) - 1] == 'rel="next"':
                res = _append_orders(link[len(link) - 2][1:-1])
            else:
                break

        return Table(orders)

    def get_query_url(self, query_date=None, since_id=None, table_name=None, count=True):
        """
        Get the URL of a Shopify API request

        `Args:`
            query_date: str
                Filter query by a date that rows were created. Format: yyyy-mm-dd. This filter
                is ignored if value is None.
            since_id: str
                Filter query by a minimum ID. This filter is ignored if value is None.
            table_name: str
                The name of the Shopify table to query.
            count: bool
                True if refund should be included in Table, False otherwise.
        `Returns:`
            str
        """
        filters = 'limit=250&status=any'

        if count:
            table = table_name + '/count.json'
        else:
            table = table_name + '.json'

        if query_date:
            # Specific date if provided
            query_date = datetime.strptime(query_date, "%Y-%m-%d")
            max_date = query_date + timedelta(days=1)
            filters += '&created_at_min={}&created_at_max={}'.format(query_date.isoformat(),
                                                                     max_date.isoformat())
        elif since_id:
            # Since ID if provided
            filters += '&since_id=%s' % since_id

        return 'https://%s.myshopify.com/admin/api/%s/%s?%s' % (
            self.subdomain,
            self.api_version,
            table,
            filters
        )

    @classmethod
    def load_to_table(cls, subdomain=None, password=None, api_key=None, api_version=None,
                      query_date=None, since_id=None, completed=True):
        """
        Fast classmethod so you can get the data all at once:
        tabledata = Shopify.load_to_table(subdomain='myorg', password='abc123',
                                         api_key='abc123', api_version='2020-10',
                                         query_date='2020-10-20', since_id='8414',
                                         completed=True)
        This instantiates the class and makes the appropriate query type to Shopify's orders
        table based on which arguments are supplied.

        `Args:`
            subdomain: str
                The Shopify subdomain (e.g. ``myorg`` for myorg.myshopify.com).
            password: str
                The Shopify account password.
            api_key: str
                The Shopify account API key.
            api_version: str
                The Shopify API version.
            query_date: str
                Filter query by a date that rows were created. Format: yyyy-mm-dd. This filter
                is ignored if value is None.
            since_id: str
                Filter query by a minimum ID. This filter is ignored if value is None.
            completed: bool
                True if only getting completed orders, False otherwise.
                value as value
        `Returns:`
            Table Class
        """
        return cls(subdomain, password, api_key, api_version).get_orders(query_date, since_id,
                                                                         completed)



class FlagIDs:
    """A class to access the FlagIDs PDI API endpoint."""

    def __init__(self):
        flags_endpoint = "/flags"
        flag_ids_endpoint = "/flag_ids"
        self.url_flags = self.base_url + flags_endpoint
        self.url_flag_ids = self.base_url + flag_ids_endpoint

        super().__init__()

    def get_flag_ids(self, limit=None):
        """Get a list of flag ids.

        `Args:`
            limit: int
                Specify limit to return.

        `Returns:`
            parsons.Table
                A Parsons table of all the data.
        """
        return self._request(self.url_flag_ids, limit=limit)

    def get_flag_id(self, id):
        """Get a specified flag id.

        `Args:`
            id: str
                The flag id identifier.

        `Returns:`
            dict
                FlagID object.
        """
        return self._request(f"{self.url_flag_ids}/{id}")

    def create_flag_id(self, flag_id, is_default, flag_description=None,
                       compile=None):
        """Save a new flag id.

        `Args:`
            flag_id: str
                The flag id type. One of: "AMM", "BNH", "BNM", "DEAD", "DNC",
                "DNR", "ENDR", "GTD", "HH", "L2VT", "LBO", "LM", "LO", "LS",
                "LSD", "LSR", "MAYBE", "MOV", "NAH", "NO", "REF", "SO", "SS",
                "SUP", "U", "UL2VT", "VL2VT", "VOL", "VTD".
            is_default: bool
                The default.
            flag_description: str
                (Optional) The flag id description.
            compile: str
                 (Optional) The compile.


        `Returns:`
            str
                The identifier for the new flag id.
        """
        payload = {
            "flagId": flag_id,
            "flagIdDescription": flag_description,
            "compile": compile,
            "isDefault": is_default,
        }
        data = self._request(
            self.url_flag_ids, req_type="POST", post_data=payload)

        return data["id"]

    def delete_flag_id(self, id):
        """Delete a flag id.

        NOTE: The function returns True (even if the id doesn't exist) unless
        there is an error.

        `Args:`
            id: str
                The flag id identifier.

        `Returns:`
            bool
                True if the operation is successful.
        """
        self._request(f"{self.url_flag_ids}/{id}", req_type="DELETE")

        return True

    def update_flag_id(self, id, flag_id, is_default, flag_description=None,
                       compile=None):
        """Update a flag id.

        `Args:`
            id: str
                The flag id identifier.
            flag_id: str
                The flag id type. One of: "AMM", "BNH", "BNM", "DEAD", "DNC",
                "DNR", "ENDR", "GTD", "HH", "L2VT", "LBO", "LM", "LO", "LS",
                "LSD", "LSR", "MAYBE", "MOV", "NAH", "NO", "REF", "SO", "SS",
                "SUP", "U", "UL2VT", "VL2VT", "VOL", "VTD".
            is_default: bool
                The default.
            flag_description: str
                (Optional) The flag id description.
            compile: str
                 (Optional) The compile.

        `Returns:`
            str
                The identifier for the udpated flag id.
        """
        payload = {
            "flagId": flag_id,
            "flagIdDescription": flag_description,
            "compile": compile,
            "isDefault": is_default,
        }
        data = self._request(
            f"{self.url_flag_ids}/{id}", req_type="PUT", post_data=payload)

        return data["id"]



import os
import unittest
from unittest import mock
import requests_mock
from parsons.sisense.sisense import Sisense

from test.test_sisense.test_data import ENV_PARAMETERS, \
    TEST_PUBLISH_SHARED_DASHBOARD, TEST_LIST_SHARED_DASHBOARDS, TEST_DELETE_SHARED_DASHBOARD


class TestSisense(unittest.TestCase):

    def setUp(self):
        self.sisense = Sisense(site_name='my_site_name', api_key='my_api_key')

    @mock.patch.dict(os.environ, ENV_PARAMETERS)
    def test_init(self):
        sisense = Sisense()
        self.assertEqual(sisense.site_name, 'my_site_name')
        self.assertEqual(sisense.api_key, 'my_api_key')
        self.assertEqual(sisense.api.uri, 'https://app.periscopedata.com/api/v1/')
        self.assertEqual(sisense.api.headers['HTTP-X-PARTNER-AUTH'], 'my_site_name:my_api_key')

    @requests_mock.Mocker()
    def test_publish_shared_dashboard(self, m):
        m.post(f'{self.sisense.uri}shared_dashboard/create', json=TEST_PUBLISH_SHARED_DASHBOARD)
        self.assertEqual(self.sisense.publish_shared_dashboard(dashboard_id='1234'), TEST_PUBLISH_SHARED_DASHBOARD) # noqa

    @requests_mock.Mocker()
    def test_list_shared_dashboards(self, m):
        m.post(f'{self.sisense.uri}shared_dashboard/list', json=TEST_LIST_SHARED_DASHBOARDS)
        self.assertEqual(self.sisense.list_shared_dashboards(dashboard_id='1234'), TEST_LIST_SHARED_DASHBOARDS)  # noqa

    @requests_mock.Mocker()
    def test_delete_shared_dashboard(self, m):
        m.post(f'{self.sisense.uri}shared_dashboard/delete', json=TEST_DELETE_SHARED_DASHBOARD)
        self.assertEqual(self.sisense.delete_shared_dashboard(token='abc'), TEST_DELETE_SHARED_DASHBOARD)  # noqa



"""NGPVAN Target Endpoints"""

from parsons.etl.table import Table
import logging
import json
import requests

logger = logging.getLogger(__name__)


class TargetsFailed(Exception):
    pass


class Targets(object):

    def __init__(self, van_connection):

        self.connection = van_connection

    def obj_dict(obj):
        return obj.__dict__

    def get_targets(self):
        """
        Get targets.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self.connection.get_request('targets'))
        logger.info(f'Found {tbl.num_rows} targets.')
        return tbl

    def get_target(self, target_id):
        """
        Get a single target.

        `Args:`
            target_id : int
                The target id.
        `Returns:`
            dict
                The target
        """

        r = self.connection.get_request(f'targets/{target_id}')
        logger.info(f'Found target {target_id}.')
        return r

    def get_target_export(self, export_job_id):
        """
        Get specific target export job id's status.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        response = self.connection.get_request(f'targetExportJobs/{export_job_id}')
        json_string = json.dumps(response)
        json_obj = json.loads(json_string)
        for i in json_obj:
            job_status = i['jobStatus']
        if job_status == 'Complete':
            for j in json_obj:
                csv = j['file']['downloadUrl']
                response_csv = requests.get(csv)
                return Table.from_csv_string(response_csv.text)
        elif job_status == 'Pending' or job_status == 'InProcess':
            logger.info(f'Target export job is pending or in process for {export_job_id}.')
        else:
            raise TargetsFailed(f'Target export failed for {export_job_id}')

    def create_target_export(self, target_id, webhook_url=None):
        """
        Create new target export job

        `Args:`
            target_id : int
                The target id the export job is creating for.
        `Returns:`
            dict
                The target export job ID
        """
        target_export = {
                        'targetId': target_id
                        }

        r = self.connection.post_request(f'targetExportJobs', json=target_export)
        logger.info(f'Created new target export job for {target_id}.')
        return r



import unittest
import os
import requests_mock
import unittest.mock as mock
from parsons.ngpvan.van import VAN
from parsons.etl.table import Table
from test.utils import assert_matching_tables
from parsons.utilities import cloud_storage

os.environ['VAN_API_KEY'] = 'SOME_KEY'


class TestBulkImport(unittest.TestCase):

    def setUp(self):

        self.van = VAN(os.environ['VAN_API_KEY'],
                       db="MyVoters", raise_for_status=False)

    def tearDown(self):

        pass

    @requests_mock.Mocker()
    def test_get_bulk_import_resources(self, m):

        json = ['Contacts', 'Contributions',
                'ActivistCodes', 'ContactsActivistCodes']

        m.get(self.van.connection.uri + 'bulkImportJobs/resources', json=json)

        self.assertEqual(self.van.get_bulk_import_resources(), json)

    @requests_mock.Mocker()
    def test_get_bulk_import_job(self, m):

        json = {'id': 59,
                'status': 'InProgress',
                'resourceType': 'ContactsActivistCodes',
                'webhookUrl': None,
                'resultFileSizeLimitKb': 5000,
                'errors': [],
                'resultFiles': []}

        m.get(self.van.connection.uri + 'bulkImportJobs/53407', json=json)

        self.assertEqual(self.van.get_bulk_import_job(53407), json)

    @requests_mock.Mocker()
    def test_get_bulk_import_mapping_types(self, m):

        m.get(self.van.connection.uri + 'bulkImportMappingTypes', json=mapping_type)

        assert_matching_tables(self.van.get_bulk_import_mapping_types(), Table(mapping_type))

    @requests_mock.Mocker()
    def test_get_bulk_import_mapping_type(self, m):

        m.get(self.van.connection.uri + 'bulkImportMappingTypes/ActivistCode', json=mapping_type)

        self.assertEqual(self.van.get_bulk_import_mapping_type('ActivistCode'), mapping_type)

    @requests_mock.Mocker()
    def test_post_bulk_import(self, m):

        # Mock Cloud Storage
        cloud_storage.post_file = mock.MagicMock()
        cloud_storage.post_file.return_value = 'https://s3.com/my_file.zip'

        tbl = Table([['Vanid', 'ActivistCodeID'], [1234, 345345]])

        m.post(self.van.connection.uri + 'bulkImportJobs', json={'jobId': 54679})

        r = self.van.post_bulk_import(tbl,
                                      'S3',
                                      'ContactsActivistCodes',
                                      [{"name": "ActivistCode"}],
                                      'Activist Code Upload',
                                      bucket='my-bucket')

        self.assertEqual(r, 54679)

    @requests_mock.Mocker()
    def test_bulk_apply_activist_codes(self, m):

        # Mock Cloud Storage
        cloud_storage.post_file = mock.MagicMock()
        cloud_storage.post_file.return_value = 'https://s3.com/my_file.zip'

        tbl = Table([['Vanid', 'ActivistCodeID'], [1234, 345345]])

        m.post(self.van.connection.uri + 'bulkImportJobs', json={'jobId': 54679})

        job_id = self.van.bulk_apply_activist_codes(tbl, url_type="S3", bucket='my-bucket')

        self.assertEqual(job_id, 54679)


mapping_type = {
    'name': 'ActivistCode',
    'displayName': 'Apply Activist Code',
    'allowMultipleMode': 'Multiple',
    'resourceTypes': ['Contacts', 'ContactsActivistCodes'],
    'fields': [
        {
            'name': 'ActivistCodeID',
            'description': 'Activist Code ID',
            'hasPredefinedValues': True,
            'isRequired': True,
            'canBeMappedToColumn': True,
            'canBeMappedByName': True,
            'parents': None
        }, {
            'name': 'CanvassedBy',
            'description': 'Recruited By, Must be a Valid User ID',
            'hasPredefinedValues': False,
            'isRequired': False,
            'canBeMappedToColumn': True,
            'canBeMappedByName': True,
            'parents': None
        }, {
            'name': 'DateCanvassed',
            'description': 'Contacted When',
            'hasPredefinedValues': False,
            'isRequired': False,
            'canBeMappedToColumn': True,
            'canBeMappedByName': True,
            'parents': [{
                'parentFieldName': 'CanvassedBy',
                'limitedToParentValues': None
            }]
        }, {
            'name': 'ContactTypeID',
            'description': 'Contacted How',
            'hasPredefinedValues': True,
            'isRequired': False,
            'canBeMappedToColumn': True,
            'canBeMappedByName': True,
            'parents': [{
                'parentFieldName': 'CanvassedBy',
                'limitedToParentValues': None
            }]
        }
    ]
}



class Questions:
    """A class to access the Questions PDI API endpoint."""

    def __init__(self):
        questions_endpoint = "/questions"
        self.url_questions = self.base_url + questions_endpoint

        super().__init__()

    def get_questions(self, limit=None):
        """Get a list of Questions.

        `Args:`
            limit: int
                Specify limit to return.

        `Returns:`
            parsons.Table
                A Parsons table of all the data.
        """

        return self._request(self.url_questions, limit=limit)



import csv
from io import TextIOWrapper, BytesIO, StringIO
import logging
import sys
import traceback
import time

from parsons.aws.aws_async import get_func_task_path, import_and_get_task, run as maybe_async_run
from parsons.aws.s3 import S3
from parsons.etl.table import Table
from parsons.utilities.check_env import check

logger = logging.getLogger(__name__)


class DistributeTaskException(Exception):
    pass


class TestStorage:

    def __init__(self):
        self.data = {}

    def put_object(self, bucket, key, object_bytes):
        self.data[key] = object_bytes

    def get_range(self, bucket, key, rangestart, rangeend):
        return self.data[key][rangestart:rangeend]


class S3Storage:
    """
    These methods are pretty specialized, so we keep them
    inside this file rather than s3.py
    """

    def __init__(self):
        self.s3 = S3()

    def put_object(self, bucket, key, object_bytes, **kwargs):
        return self.s3.client.put_object(Bucket=bucket, Key=key, Body=object_bytes, **kwargs)

    def get_range(self, bucket, key, rangestart, rangeend):
        """
        Gets an explicit byte-range of an S3 file
        """
        # bytes is INCLUSIVE for the rangeend parameter, unlike python
        # so e.g. while python returns 2 bytes for data[2:4]
        # Range: bytes=2-4 will return 3!! So we subtract 1
        response = self.s3.client.get_object(
            Bucket=bucket, Key=key,
            Range='bytes={}-{}'.format(rangestart, rangeend - 1))
        return response['Body'].read()


FAKE_STORAGE = TestStorage()
S3_TEMP_KEY_PREFIX = "Parsons_DistributeTask"


def distribute_task_csv(csv_bytes_utf8, func_to_run, bucket,
                        header=None,
                        func_kwargs=None,
                        func_class=None,
                        func_class_kwargs=None,
                        catch=False,
                        group_count=100,
                        storage='s3'):
    """
    The same as distribute_task, but instead of a table, the
    first argument is bytes of a csv encoded into utf8.
    This function is used by distribute_task() which you should use instead.
    """
    global FAKE_STORAGE
    func_name = get_func_task_path(func_to_run, func_class)
    row_chunks = csv_bytes_utf8.split(b'\n')
    cursor = 0
    row_ranges = []
    # gather start/end bytes for each row
    for rowc in row_chunks:
        rng = [cursor]
        cursor = cursor + len(rowc) + 1  # +1 is the \n character
        rng.append(cursor)
        row_ranges.append(rng)

    # group the rows and get start/end bytes for each group
    group_ranges = []
    # table csv writer appends a terminal \r\n, so we do len-1
    for grpstep in range(0, len(row_ranges) - 1, group_count):
        end = min(len(row_ranges) - 1, grpstep + group_count - 1)
        group_ranges.append((row_ranges[grpstep][0], row_ranges[end][1]))

    # upload data
    filename = hash(time.time())
    storagekey = f"{S3_TEMP_KEY_PREFIX}/{filename}.csv"
    groupcount = len(group_ranges)
    logger.debug(f'distribute_task_csv storagekey {storagekey} w/ {groupcount} groups')

    response = None
    if storage == 's3':
        response = S3Storage().put_object(bucket, storagekey, csv_bytes_utf8)
    else:
        response = FAKE_STORAGE.put_object(bucket, storagekey, csv_bytes_utf8)

    # start processes
    results = [
        maybe_async_run(
            process_task_portion,
            [bucket, storagekey, grp[0], grp[1], func_name, header,
             storage, func_kwargs, catch, func_class_kwargs],
            # if we are using local storage, then it must be run locally, as well
            # (good for testing/debugging)
            remote_aws_lambda_function_name='FORCE_LOCAL' if storage == 'local' else None
        )
        for grp in group_ranges]
    return {'DEBUG_ONLY': 'results may vary depending on context/platform',
            'results': results,
            'put_response': response}


def distribute_task(table, func_to_run,
                    bucket=None,
                    func_kwargs=None,
                    func_class=None,
                    func_class_kwargs=None,
                    catch=False,
                    group_count=100,
                    storage='s3'):
    """
    Distribute processing rows in a table across multiple AWS Lambda invocations.

    `Args:`
        table: Parsons Table
           Table of data you wish to distribute processing across Lambda invocations
           of `func_to_run` argument.
        func_to_run: function
           The function you want to run whose
           first argument will be a subset of table
        bucket: str
           The bucket name to use for s3 upload to process the whole table
           Not required if you set environment variable ``S3_TEMP_BUCKET``
        func_kwargs: dict
           If the function has other arguments to pass along with `table`
           then provide them as a dict here. They must all be JSON-able.
        func_class: class
           If the function is a classmethod or function on a class,
           then pass the pure class here.
           E.g. If you passed `ActionKit.bulk_upload_table`,
           then you would pass `ActionKit` here.
        func_class_kwargs: dict
           If it is a class function, and the class must be instantiated,
           then pass the kwargs to instantiate the class here.
           E.g. If you passed `ActionKit.bulk_upload_table` as the function,
           then you would pass {'domain': ..., 'username': ... etc} here.
           This must all be JSON-able data.
        catch: bool
           Lambda will retry running an event three times if there's an
           exception -- if you want to prevent this, set `catch=True`
           and then it will catch any errors and stop retries.
           The error will be in CloudWatch logs with string "Distribute Error"
           This might be important if row-actions are not idempotent and your
           own function might fail causing repeats.
        group_count: int
           Set this to how many rows to process with each Lambda invocation (Default: 100)
        storage: str
           Debugging option: Defaults to "s3". To test distribution locally without s3,
           set to "local".
    `Returns:`
        Debug information -- do not rely on the output, as it will change
        depending on how this method is invoked.
    """
    if storage not in ('s3', 'local'):
        raise DistributeTaskException(f'storage argument must be s3 or local')
    bucket = check('S3_TEMP_BUCKET', bucket)
    csvdata = StringIO()
    outcsv = csv.writer(csvdata)
    outcsv.writerows(table.table.data())
    return distribute_task_csv(csvdata.getvalue().encode('utf-8-sig'),
                               func_to_run,
                               bucket,
                               header=table.columns,
                               func_kwargs=func_kwargs,
                               func_class=func_class,
                               func_class_kwargs=func_class_kwargs,
                               catch=catch,
                               group_count=group_count,
                               storage=storage)


def process_task_portion(bucket, storagekey, rangestart, rangeend, func_name, header,
                         storage='s3', func_kwargs=None, catch=False,
                         func_class_kwargs=None):
    global FAKE_STORAGE

    logger.debug(f'process_task_portion func_name {func_name}, '
                 f'storagekey {storagekey}, byterange {rangestart}-{rangeend}')
    func = import_and_get_task(func_name, func_class_kwargs)
    if storage == 's3':
        filedata = S3Storage().get_range(bucket, storagekey, rangestart, rangeend)
    else:
        filedata = FAKE_STORAGE.get_range(bucket, storagekey, rangestart, rangeend)

    lines = list(csv.reader(TextIOWrapper(BytesIO(filedata), encoding='utf-8-sig')))
    table = Table([header] + lines)
    if catch:
        try:
            func(table, **func_kwargs)
        except Exception:
            # In Lambda you can search for '"Distribute Error"' in the logs
            type_, value_, traceback_ = sys.exc_info()
            err_traceback_str = '\n'.join(traceback.format_exception(type_, value_, traceback_))
            return {'Exception': 'Distribute Error',
                    'error': err_traceback_str,
                    'rangestart': rangestart,
                    'rangeend': rangeend,
                    'func_name': func_name,
                    'bucket': bucket,
                    'storagekey': storagekey}
    else:
        func(table, **func_kwargs)



import unittest
import os
import requests_mock
from parsons.ngpvan import VAN
from parsons.etl.table import Table
from test.utils import validate_list, assert_matching_tables

os.environ['VAN_API_KEY'] = 'SOME_KEY'


class TestTargets(unittest.TestCase):

    mock_data = mock_data = (
        '12827,Volunteer Recruitment Tiers,Tier,109957740\n'
        '12827,Volunteer Recruitment Tiers,Tier,109957754')
    mock_result = Table([
        ('12827', 'Volunteer Recruitment Tiers', 'Tier', '109957740'),
        ('12827', 'Volunteer Recruitment Tiers', 'Tier', '109957754')])

    def setUp(self):

        self.van = VAN(os.environ['VAN_API_KEY'], db="MyVoters", raise_for_status=False)

    def tearDown(self):

        pass

    @requests_mock.Mocker()
    def test_get_targets(self, m):

        # Create response
        json = {u'count': 2, u'items':
                [{u'targetId': 12827,
                  u'type': u'TEST CODE',
                  u'name': u'TEST CODE',
                  u'description': None,
                  u'points': 20,
                  u'areSubgroupsSticky': False,
                  u'status': u'Active',
                  u'subgroups': None,
                  u'markedSubgroup': None}],
                u'nextPageLink': None}

        m.get(self.van.connection.uri + 'targets', json=json)

        # Expected Structure
        expected = ['targetId', 'type', 'name', 'description',
                    'points', 'areSubgroupsSticky', 'status', 'subgroups', 'markedSubgroup']

        # Assert response is expected structure
        self.assertTrue(validate_list(expected, self.van.get_targets()))

        # To Do: Test what happens when it doesn't find any targets

    @requests_mock.Mocker()
    def test_get_target(self, m):

        # Create response
        json = {u'targetId': 15723,
                u'name': u'Mail_VR_Chase',
                u'type': u'Dynamic',
                u'description': None,
                u'points': 15,
                u'areSubgroupsSticky': False,
                u'status': u'Active',
                u'subgroups':
                [{u'targetId': 12827,
                  u'fullName': u'April_VR_Chase Calls',
                  u'name': u'April_Chase_20',
                  u'subgroupId': 46803,
                  u'isAssociatedWithBadges': True}],
                u'markedSubgroup': None}

        m.get(self.van.connection.uri + 'targets/15723', json=json)

        self.assertEqual(json, self.van.get_target(15723))

    @requests_mock.Mocker()
    def test_create_target_export(self, m):

        export_job_id = '{"exportJobId": "455961790"}'
        target_id = 12827

        m.post(self.van.connection.uri + 'targetExportJobs', json=export_job_id, status_code=204)

        # Test that it doesn't throw and error
        r = self.van.create_target_export(target_id, webhook_url=None)

        self.assertEqual(r, export_job_id)

    @requests_mock.Mocker()
    def test_get_target_export(self, m):

        export_job_id = 455961790
        json = [{
            "targetId": 12827,
            "file": {
                "downloadUrl": (
                    "https://ngpvan.blob.core.windows.net/"
                    "target-export-files/TargetExport_455961790.csv"),
                "dateExpired": "null",
                "recordCount": 1016883},
            "webhookUrl": "null",
            "exportJobId": 455961790,
            "jobStatus": "Complete"}]

        download_url = (
            'https://ngpvan.blob.core.windows.net/target-export-files/TargetExport_455961790.csv')

        m.post(self.van.connection.uri + 'targetExportJobs', json=export_job_id, status_code=204)
        m.get(self.van.connection.uri + 'targetExportJobs/455961790', json=json)
        m.get(download_url, text=self.mock_data)
        assert_matching_tables(self.van.get_target_export(export_job_id),
                               self.mock_result)



import logging
import os
from urllib.parse import urlparse

from azure.core.exceptions import ResourceNotFoundError
from azure.storage.blob import BlobServiceClient, ContentSettings, generate_blob_sas

from parsons.utilities import check_env, files

logger = logging.getLogger(__name__)


class AzureBlobStorage(object):
    """
    Instantiate AzureBlobStorage Class for a given Azure storage account.

    `Args:`
        account_name: str
            The name of the Azure storage account to use. Not required if ``AZURE_ACCOUNT_NAME``
            environment variable is set, or if ``account_url`` is supplied.
        credential: str
            An account shared access key with access to the Azure storage account, an SAS token
            string, or an instance of a TokenCredentials class. Not required if ``AZURE_CREDENTIAL``
            environment variable is set.
        account_domain: str
            The domain of the Azure storage account, defaults to "blob.core.windows.net".
            Not required if ``AZURE_ACCOUNT_DOMAIN`` environment variable is set or if
            ``account_url`` is supplied.
        account_url: str
            The account URL for the Azure storage account including the account name and domain.
            Not required if ``AZURE_ACCOUNT_URL`` environment variable is set.
    `Returns:`
        `AzureBlobStorage`
    """

    def __init__(self, account_name=None, credential=None, account_domain='blob.core.windows.net',
                 account_url=None):
        self.account_url = os.getenv('AZURE_ACCOUNT_URL', account_url)
        self.credential = check_env.check('AZURE_CREDENTIAL', credential)
        if not self.account_url:
            self.account_name = check_env.check('AZURE_ACCOUNT_NAME', account_name)
            self.account_domain = check_env.check('AZURE_ACCOUNT_DOMAIN', account_domain)
            self.account_url = f'https://{self.account_name}.{self.account_domain}/'
        else:
            if not self.account_url.startswith('http'):
                self.account_url = f'https://{self.account_url}'
            # Update the account name and domain if a URL is supplied
            parsed_url = urlparse(self.account_url)
            self.account_name = parsed_url.netloc.split(".")[0]
            self.account_domain = ".".join(parsed_url.netloc.split(".")[1:])
        self.client = BlobServiceClient(account_url=self.account_url, credential=self.credential)

    def list_containers(self):
        """
        Returns a list of container names for the storage account

        `Returns:`
            list[str]
                List of container names
        """

        container_names = [container.name for container in self.client.list_containers()]
        logger.info(f'Found {len(container_names)} containers.')
        return container_names

    def container_exists(self, container_name):
        """
        Verify that a container exists within the storage account

        `Args:`
            container_name: str
                The name of the container
        `Returns:`
            bool
        """

        container_client = self.get_container(container_name)
        try:
            container_client.get_container_properties()
            logger.info(f'{container_name} exists.')
            return True
        except ResourceNotFoundError:
            logger.info(f'{container_name} does not exist.')
            return False

    def get_container(self, container_name):
        """
        Returns a container client

        `Args:`
            container_name: str
                The name of the container
        `Returns:`
            `ContainerClient`
        """

        logger.info(f'Returning {container_name} container client')
        return self.client.get_container_client(container_name)

    def create_container(self, container_name, metadata=None, public_access=None, **kwargs):
        """
        Create a container

        `Args:`
            container_name: str
                The name of the container
            metadata: Optional[dict[str, str]]
                A dict with metadata to associated with the container.
            public_access: Optional[Union[PublicAccess, str]]
                Settings for public access on the container, can be 'container' or 'blob' if not
                ``None``
            kwargs:
                Additional arguments to be supplied to the Azure Blob Storage API. See `Azure Blob
                Storage SDK documentation <https://docs.microsoft.com/en-us/python/api/azure-storage-blob/azure.storage.blob.blobserviceclient?view=azure-python#create-container-name--metadata-none--public-access-none----kwargs->`_
                for more info.
        `Returns:`
            `ContainerClient`
        """  # noqa

        container_client = self.client.create_container(
            container_name, metadata=metadata, public_access=public_access, **kwargs
        )
        logger.info(f'Created {container_name} container.')
        return container_client

    def delete_container(self, container_name):
        """
        Delete a container.

        `Args:`
            container_name: str
                The name of the container
        `Returns:`
            ``None``
        """

        self.client.delete_container(container_name)
        logger.info(f'{container_name} container deleted.')

    def list_blobs(self, container_name, name_starts_with=None):
        """
        List all of the names of blobs in a container

        `Args:`
            container_name: str
                The name of the container
            name_starts_with: Optional[str]
                A prefix to filter blob names
        `Returns:`
            list[str]
                A list of blob names
        """

        container_client = self.get_container(container_name)
        blobs = [
            blob for blob in container_client.list_blobs(name_starts_with=name_starts_with)
        ]
        logger.info(f'Found {len(blobs)} blobs in {container_name} container.')
        return blobs

    def blob_exists(self, container_name, blob_name):
        """
        Verify that a blob exists in the specified container

        `Args:`
            container_name: str
                The container name
            blob_name: str
                The blob name
        `Returns:`
            bool
        """

        blob_client = self.get_blob(container_name, blob_name)
        try:
            blob_client.get_blob_properties()
            logger.info(f'{blob_name} exists in {container_name} container.')
            return True
        except ResourceNotFoundError:
            logger.info(f'{blob_name} does not exist in {container_name} container.')
            return False

    def get_blob(self, container_name, blob_name):
        """
        Get a blob object

        `Args:`
            container_name: str
                The container name
            blob_name: str
                The blob name
        `Returns:`
            `BlobClient`
        """

        blob_client = self.client.get_blob_client(container_name, blob_name)
        logger.info(f'Got {blob_name} blob from {container_name} container.')
        return blob_client

    def get_blob_url(self, container_name, blob_name, account_key=None, permission=None,
                     expiry=None, start=None):
        """
        Get a URL with a shared access signature for a blob

        `Args:`
            container_name: str
                The container name
            blob_name: str
                The blob name
            account_key: Optional[str]
                An account shared access key for the storage account. Will default to the key used
                on initialization if one was provided as the credential, but required if it was not.
            permission: Optional[Union[BlobSasPermissions, str]]
                Permissions associated with the blob URL. Can be either a BlobSasPermissions object
                or a string where 'r', 'a', 'c', 'w', and 'd' correspond to read, add, create,
                write, and delete permissions respectively.
            expiry: Optional[Union[datetime, str]]
                The datetime when the URL should expire. Defaults to UTC.
            start: Optional[Union[datetime, str]]
                The datetime when the URL should become valid. Defaults to UTC. If it is ``None``,
                the URL becomes active when it is first created.
        `Returns:`
            str
                URL with shared access signature for blob
        """

        if not account_key:
            if not self.credential:
                raise ValueError(
                    f'An account shared access key must be provided if it was not on initialization'
                )
            account_key = self.credential

        sas = generate_blob_sas(
            self.account_name,
            container_name,
            blob_name,
            account_key=account_key,
            permission=permission,
            expiry=expiry,
            start=start,
        )
        return f'{self.account_url}/{container_name}/{blob_name}?sas={sas}'

    def _get_content_settings_from_dict(self, kwargs_dict):
        """
        Removes any keys for ``ContentSettings`` from a dict and returns a tuple of the generated
        settings or ``None`` and a dict with the settings keys removed.

        `Args:`
            kwargs_dict: dict
                A dict which should be processed and may have keys for ``ContentSettings``
        `Returns:`
            Tuple[Optional[ContentSettings], dict]
                Any created settings or ``None`` and the dict with settings keys remvoed
        """

        kwargs_copy = {**kwargs_dict}
        content_settings = None
        content_settings_dict = {}
        content_settings_keys = [
            'content_type', 'content_encoding', 'content_language', 'content_disposition',
            'cache_control', 'content_md5'
        ]
        kwarg_keys = list(kwargs_copy.keys())
        for key in kwarg_keys:
            if key in content_settings_keys:
                content_settings_dict[key] = kwargs_copy.pop(key)
        if content_settings_dict:
            content_settings = ContentSettings(**content_settings_dict)

        return content_settings, kwargs_copy

    def put_blob(self, container_name, blob_name, local_path, **kwargs):
        """
        Puts a blob (aka file) in a bucket

        `Args:`
            container_name: str
                The name of the container to store the blob
            blob_name: str
                The name of the blob to be stored
            local_path: str
                The local path of the file to upload
            kwargs:
                Additional arguments to be supplied to the Azure Blob Storage API. See `Azure Blob
                Storage SDK documentation <https://docs.microsoft.com/en-us/python/api/azure-storage-blob/azure.storage.blob.blobclient?view=azure-python#upload-blob-data--blob-type--blobtype-blockblob---blockblob----length-none--metadata-none----kwargs->`_
                for more info. Any keys that belong to the ``ContentSettings`` object will be
                provided to that class directly.
        `Returns:`
            `BlobClient`
        """  # noqa

        blob_client = self.get_blob(container_name, blob_name)

        # Move all content_settings keys into a ContentSettings object
        content_settings, kwargs_dict = self._get_content_settings_from_dict(kwargs)

        with open(local_path, 'rb') as f:
            data = f.read()

        blob_client = blob_client.upload_blob(
            data,
            overwrite=True,
            content_settings=content_settings,
            **kwargs_dict,
        )
        logger.info(f'{blob_name} blob put in {container_name} container')

        # Return refreshed BlobClient object
        return self.get_blob(container_name, blob_name)

    def download_blob(self, container_name, blob_name, local_path=None):
        """
        Downloads a blob from a container into the specified file path or a temporary file path

        `Args:`
            container_name: str
                The container name
            blob_name: str
                The blob name
            local_path: Optional[str]
                The local path where the file will be downloaded. If not specified, a temporary
                file will be created and returned, and that file will be removed automatically
                when the script is done running.
        `Returns:`
            str
                The path of the downloaded file
        """

        if not local_path:
            local_path = files.create_temp_file_for_path('TEMPFILEAZURE')

        blob_client = self.get_blob(container_name, blob_name)

        logger.info(f'Downloading {blob_name} blob from {container_name} container.')
        with open(local_path, 'wb') as f:
            blob_client.download_blob().readinto(f)
        logger.info(f'{blob_name} blob saved to {local_path}.')

        return local_path

    def delete_blob(self, container_name, blob_name):
        """
        Delete a blob in a specified container.

        `Args:`
            container_name: str
                The container name
            blob_name: str
                The blob name
        `Returns:`
            ``None``
        """

        blob_client = self.get_blob(container_name, blob_name)
        blob_client.delete_blob()
        logger.info(f'{blob_name} blob in {container_name} container deleted.')

    def upload_table(self, table, container_name, blob_name, data_type='csv', **kwargs):
        """
        Load the data from a Parsons table into a blob.

        `Args:`
            table: obj
                A :ref:`parsons-table`
            container_name: str
                The container name to upload the data into
            blob_name: str
                The blob name to upload the data into
            data_type: str
                The file format to use when writing the data. One of: `csv` or `json`
            kwargs:
                Additional keyword arguments to supply to ``put_blob``
        `Returns:`
            `BlobClient`
        """

        if data_type == 'csv':
            local_path = table.to_csv()
            content_type = 'text/csv'
        elif data_type == 'json':
            local_path = table.to_json()
            content_type = 'application/json'
        else:
            raise ValueError(f'Unknown data_type value ({data_type}): must be one of: csv or json')

        return self.put_blob(
            container_name, blob_name, local_path, content_type=content_type, **kwargs
        )
